# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A pyspiel config for a debate with randomly named debaters.
"""

import collections

from ml_collections import config_dict

from open_spiel.python.games.chat_games.envs.base_envs import debate_with_style_info as env_debate_with_style_info
from open_spiel.python.games.chat_games.envs.observations import summary_debate
from open_spiel.python.games.chat_games.envs.observations import utils as obs_utils
from open_spiel.python.games.chat_games.envs.payoffs import debate as payoffs_debate
from open_spiel.python.games.chat_games.envs.scenarios.actions import arguments
from open_spiel.python.games.chat_games.envs.scenarios.domains import debate as scenario_debate
from open_spiel.python.games.chat_games.envs.scenarios.players import names as names_debate


def get_config():
  """Get configuration for chat game."""
  config = config_dict.ConfigDict()

  num_players = 2

  observations = [
      obs_utils.Observation(summary_debate.PREFIX, summary_debate.POSTFIX)
      for _ in range(num_players)
  ]

  header = env_debate_with_style_info.HEADER

  payoffs = [payoffs_debate.PAYOFF]

  examples_names = names_debate.NAMES

  given_prompt_actions = collections.OrderedDict()
  given_prompt_actions[header.action_keys[0]] = arguments.STYLES + ['any']
  num_styles = len(arguments.STYLES) + 1

  given_private_info = collections.OrderedDict()
  given_private_info['info'] = ['Argue for the topic statement.',
                                'Argue against the topic statement.']
  given_private_info['topic'] = [scenario_debate.TOPIC_B,
                                 scenario_debate.TOPIC_B]

  scenario_a = env_debate_with_style_info.Scenario(
      '',
      'Bob',
      'Alice',
      'logos',
      scenario_debate.TOPIC_B,
      'Argue for the topic statement.')

  examples_scenarios = [scenario_a]

  llm_termination_prompt = scenario_debate.LLM_TERMINATION_PROMPT

  params = {'num_distinct_actions': num_players * num_styles,
            'num_llm_seeds': 2,
            'num_init_states': 3,
            'num_players': num_players,
            'min_utility': min([float(p.min) for p in payoffs]),
            'max_utility': max([float(p.max) for p in payoffs]),
            'num_max_replies': 1,
            'silence_logging': True}

  config.params = params

  config.game = config_dict.ConfigDict()
  config.game.observations = observations
  config.game.header = header
  config.game.payoffs = payoffs
  config.game.given_prompt_actions = given_prompt_actions
  config.game.num_names = 10
  config.game.num_private_info = (2, 2)
  config.game.examples_names = examples_names
  config.game.given_private_info = given_private_info
  config.game.examples_scenarios = examples_scenarios
  config.game.llm_list_suffix = 'Output: '
  config.game.llm_termination_prompt = llm_termination_prompt

  return config



# Copyright 2026 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements the ESCHER algorithm.

See https://arxiv.org/abs/2206.04122.

ESCHER is an unbiased model-free method that does not require any importance
sampling. Emperically, the variance of the estimated regret of ESCHER is orders
of magnitude lower than DREAM and other baselines.
"""

import logging
import typing

import numpy as np
import open_spiel
import torch

from open_spiel.python.algorithms import exploitability as _
from open_spiel.python.pytorch import deep_cfr as _

# pylint: disable=invalid-name


class Config:
  """A Config is an configuration for an Escher agent."""

  def __init__(self):
    """Initialize the configuration.

    These parameters are designed for Kuhn poker to achieve an
    exploitability around 0.05 within 100 iterations.
    """
    self.value_traversals = 512
    self.value_exploration = 0.1
    self.value_memory_capacity = int(1e6)
    self.value_net = [64]
    self.value_batch_size = 256
    self.value_batch_steps = 512
    self.value_learning_rate = 1e-3

    self.regret_traversals = 1024
    self.regret_memory_capacity = int(1e6)
    self.regret_net = [64]
    self.regret_batch_size = 256
    self.regret_batch_steps = 375
    self.regret_learning_rate = 1e-3

    self.avg_policy_memory_capacity = int(1e6)
    self.avg_policy_net = [64]
    self.avg_policy_batch_size = 256
    self.avg_policy_batch_steps = 2500
    self.avg_policy_learning_rate = 1e-3


class Agent:
  """An Agent is an ESCHER agent.

  See https://arxiv.org/abs/2206.04122.
  """

  def __init__(self, game, cfg):
    """Initialize the agent.

    Args:
        game: Openspiel game.
        cfg: (Config) configuration for the agent.
    """
    self.cfg = cfg
    self.t = 1

    # Get game state dimensions.
    state = game.new_initial_state()
    history_dim = _state_history(game.num_players(), state).shape[0]
    obs_dim = game.information_state_tensor_size()
    action_dim = game.num_distinct_actions()

    # Initialize average policy network.
    ReservoirBuffer = open_spiel.python.pytorch.deep_cfr.ReservoirBuffer
    infoset = np.zeros(obs_dim, dtype=float)
    policy = np.zeros(action_dim, dtype=float)
    scalar = np.array(0, dtype=float)
    self.avg_policy_buffer = ReservoirBuffer.init(
        cfg.avg_policy_memory_capacity,
        Behaviour(state=infoset, policy=policy, t=scalar),
    )
    MLP = open_spiel.python.pytorch.deep_cfr.MLP
    self.avg_policy_net = MLP(obs_dim, cfg.avg_policy_net, action_dim)

    # Initialize regret network.
    sr = StateRegret(state=infoset, regret=policy, mask=policy, t=scalar)
    self.regret_buffers = [
        ReservoirBuffer.init(cfg.regret_memory_capacity, sr)
        for _ in range(game.num_players())
    ]
    self.regret_nets = [
        MLP(obs_dim, cfg.regret_net, action_dim)
        for _ in range(game.num_players())
    ]

    # Initialize value network.
    history = np.zeros(history_dim, dtype=float)
    sav = StateActionValue(state=history, action=scalar, value=scalar)
    self.value_buffers = [
        ReservoirBuffer.init(cfg.value_memory_capacity, sav)
        for _ in range(game.num_players())
    ]
    self.value_nets = [
        MLP(history_dim, cfg.value_net, 1) for _ in range(game.num_players())
    ]

    self.num_touched = 0
    self.avg_policy_t = 0
    self.regret_t = 0
    self.value_t = 0

  def action_probabilities(self, state):
    """action_probabilities returns the action probabilities of state.

    Args:
        state: (pyspiel.State) The state to compute probabilities for.

    Returns:
        Action probabilities of state.
    """
    obs = np.array(state.information_state_tensor(), dtype=float)
    mask_np = np.array(state.legal_actions_mask(), dtype=int)

    with torch.no_grad():
      x = torch.from_numpy(obs).to(torch.float32)
      mask = torch.from_numpy(mask_np)

      logits = self.avg_policy_net(x)
      probs = torch.nn.functional.softmax(logits, dim=0)

      probs = torch.mul(probs, mask)
      probs = probs / torch.sum(probs)

    return probs.cpu().numpy()


def _train_avg_policy(agent):
  """Trains the average policy network."""
  num_epoch = 8
  epoch_steps = int(np.ceil(agent.cfg.avg_policy_batch_steps / num_epoch))

  buf = agent.avg_policy_buffer
  dataset = torch.utils.data.TensorDataset(
      torch.from_numpy(buf.experience.state),
      torch.from_numpy(buf.experience.policy),
      torch.from_numpy(buf.experience.t),
  )
  optimizer = torch.optim.Adam(
      agent.avg_policy_net.parameters(), lr=agent.cfg.avg_policy_learning_rate
  )

  for _ in range(num_epoch):
    agent.avg_policy_t += 1

    agent.avg_policy_net.train()
    for _ in range(epoch_steps):
      indices = np.random.choice(
          len(buf), size=(agent.cfg.avg_policy_batch_size,), replace=False
      )
      batch = Behaviour(
          state=dataset.tensors[0][indices],
          policy=dataset.tensors[1][indices],
          t=dataset.tensors[2][indices],
      )

      loss = _get_avg_policy_loss(agent, batch)
      optimizer.zero_grad()
      loss.backward()
      optimizer.step()


def _gather_regret_data(game, agent, player):
  """Gathers regret data for training."""
  for _ in range(agent.cfg.regret_traversals):
    state = game.new_initial_state()
    agent.num_touched += 1
    while not state.is_terminal():
      if state.is_chance_node():
        actions, probs = zip(*state.chance_outcomes())
        a = np.random.choice(actions, p=probs)
        state.apply_action(a)
        continue

      # Get policy.
      current_player = state.current_player()
      obs = np.array(state.information_state_tensor(), dtype=float)
      mask = np.array(state.legal_actions_mask(), dtype=int)
      policy = _match_regret(agent.regret_nets[current_player], obs, mask)

      # Add data to buffer.
      if current_player == player:
        regret = _get_regret(agent, state, policy, game.num_players())
        sr = StateRegret(state=obs, regret=regret, mask=mask, t=agent.t)
        agent.regret_buffers[player].append(sr)
      else:
        behaviour = Behaviour(state=obs, policy=policy, t=agent.t)
        agent.avg_policy_buffer.append(behaviour)

      # Update state with policy.
      if current_player == player:
        sample_policy = mask / np.sum(mask)
      else:
        sample_policy = policy
      action = np.random.choice(range(len(sample_policy)), p=sample_policy)
      state = state.child(action)
      agent.num_touched += 1


def _train_regret(cfg, agent):
  """Trains the regret network."""
  for player in range(cfg.game.num_players()):
    _train_value(cfg, agent, player)
    _gather_regret_data(cfg.game, agent, player)

    num_epoch = 8
    epoch_steps = int(np.ceil(agent.cfg.regret_batch_steps / num_epoch))
    buf = agent.regret_buffers[player]
    dataset = torch.utils.data.TensorDataset(
        torch.from_numpy(buf.experience.state),
        torch.from_numpy(buf.experience.regret),
        torch.from_numpy(buf.experience.mask),
        torch.from_numpy(buf.experience.t),
    )
    regret_net = agent.regret_nets[player]
    regret_net.reset()
    optimizer = torch.optim.Adam(
        regret_net.parameters(), lr=agent.cfg.regret_learning_rate
    )

    for _ in range(num_epoch):
      agent.regret_t += 1

      for _ in range(epoch_steps):
        indices = np.random.choice(
            len(buf), size=(agent.cfg.regret_batch_size,), replace=False
        )
        batch = StateRegret(
            state=dataset.tensors[0][indices],
            regret=dataset.tensors[1][indices],
            mask=dataset.tensors[2][indices],
            t=dataset.tensors[3][indices],
        )

        loss = _get_regret_loss(agent, player, batch)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()


def _gather_value_data(game, agent, player):
  """Gathers value data for training."""
  value_buffer = agent.value_buffers[player]
  value_buffer.clear()
  for _ in range(agent.cfg.value_traversals):
    state = game.new_initial_state()
    agent.num_touched += 1
    transitions = []
    while True:
      if state.is_chance_node():
        actions, probs = zip(*state.chance_outcomes())
        a = np.random.choice(actions, p=probs)
        state.apply_action(a)
        continue

      action, importance = -1, 1
      if not state.is_terminal():
        # Get policy.
        obs = np.array(state.information_state_tensor(), dtype=float)
        mask = np.array(state.legal_actions_mask(), dtype=int)
        regret_net = agent.regret_nets[state.current_player()]
        policy = _match_regret(regret_net, obs, mask)

        # Sample action.
        epsilon = agent.cfg.value_exploration
        uniform = mask / np.sum(mask)
        sample_policy = epsilon * uniform + (1 - epsilon) * policy
        action = np.random.choice(range(len(sample_policy)), p=sample_policy)
        importance = policy[action] / sample_policy[action]

      # Add transition.
      history = _state_history(game.num_players(), state)
      returns = np.array(state.returns(), dtype=float)
      tn = Transition(
          history=history, importance=importance, action=action, returns=returns
      )
      transitions.append(tn)

      if state.is_terminal():
        break
      state = state.child(action)
      agent.num_touched += 1

    value = np.zeros(transitions[0].returns.shape, dtype=float)
    for i in range(len(transitions) - 1, -1, -1):
      tn = transitions[i]

      value = tn.importance * (tn.returns + value)
      value_buffer.append(
          StateActionValue(
              state=tn.history, action=tn.action, value=value[player]
          )
      )


def _train_value(cfg, agent, player):
  """Trains the value network."""
  _gather_value_data(cfg.game, agent, player)

  num_epoch = 8
  epoch_steps = int(np.ceil(agent.cfg.value_batch_steps / num_epoch))
  buf = agent.value_buffers[player]
  dataset = torch.utils.data.TensorDataset(
      torch.from_numpy(buf.experience.state),
      torch.from_numpy(buf.experience.action),
      torch.from_numpy(buf.experience.value),
  )
  value_net = agent.value_nets[player]
  value_net.reset()
  optimizer = torch.optim.Adam(
      value_net.parameters(), lr=agent.cfg.value_learning_rate
  )

  for _ in range(num_epoch):
    agent.value_t += 1

    agent.value_nets[player].train()
    for _ in range(epoch_steps):
      indices = np.random.choice(
          len(buf), size=(agent.cfg.value_batch_size,), replace=False
      )
      batch = StateActionValue(
          state=dataset.tensors[0][indices],
          action=dataset.tensors[1][indices],
          value=dataset.tensors[2][indices],
      )

      loss = _get_value_loss(agent, player, batch)

      optimizer.zero_grad()
      loss.backward()
      optimizer.step()


def _get_avg_policy_loss(agent, batch):
  """Returns the loss for the average policy network."""
  x = batch.state.to(torch.float32)
  y_policy = batch.policy

  logits = agent.avg_policy_net(x)

  loss = torch.nn.functional.cross_entropy(logits, y_policy)

  # Linear CFR.
  weight = batch.t / agent.t
  loss = torch.mul(loss, weight)

  return torch.mean(loss)


def _get_regret_loss(agent, player, batch):
  """Returns the loss for the regret network."""
  x = batch.state.to(torch.float32)
  mask = batch.mask
  y_regret = batch.regret

  regret = agent.regret_nets[player](x)

  loss = torch.pow(regret - y_regret, 2)

  # Linear CFR.
  weight = batch.t / agent.t
  weight = weight.unsqueeze(-1).expand(-1, loss.shape[-1])
  loss = torch.mul(loss, weight)

  loss = torch.sum(torch.mul(loss, mask)) / torch.sum(mask)
  return loss


def _get_value_loss(agent, player, batch):
  x = batch.state.to(torch.float32)
  y_value = batch.value

  value = agent.value_nets[player](x)
  value = torch.squeeze(value, dim=[1])

  loss = torch.pow(value - y_value, 2)
  return torch.mean(loss)


def _match_regret(net, obs, mask_np):
  """Returns the policy after applying regret matching."""
  with torch.no_grad():
    x = torch.from_numpy(obs).to(torch.float32)
    regrets = net(x)
    raw_regrets = regrets.cpu().numpy()

  regrets = np.clip(raw_regrets, a_min=0, a_max=None)
  regrets = regrets * mask_np
  summed = np.sum(regrets)
  if summed > 1e-6:
    return regrets / summed

  # Just use the best regret, if regrets cannot be normalized.
  max_id, max_regret = -1, float("-inf")
  for i, m in enumerate(mask_np):
    if m == 1 and raw_regrets[i] > max_regret:
      max_id, max_regret = i, raw_regrets[i]
  policy = np.zeros(regrets.shape, dtype=regrets.dtype)
  policy[max_id] = 1
  return policy


def _get_regret(agent, state, policy, num_players):
  """Returns the regret for the current state."""
  player = state.current_player()

  mask = state.legal_actions_mask()
  children_values = np.zeros(len(mask), dtype=float)
  for a, m in enumerate(mask):
    if m == 1:
      child = state.child(a)

      with torch.no_grad():
        history = _state_history(num_players, child)
        x = torch.from_numpy(history).to(torch.float32)
        children_values[a] = agent.value_nets[player](x)

  value = np.sum(policy * children_values)
  regret = children_values - value
  return regret


class TrainConfig:
  """A TrainConfig is a configuration for the training of an Escher agent."""

  def __init__(self, game):
    """Initialize the training configuration.

    Args:
        game: Openspiel game.
    """
    self.game = game
    self.device = "cpu"
    self.iterations = 999999
    self.evaluation_interval = 1
    self.nashconv = False
    self.games_vs_random = 1000


def train(cfg, agent):
  """Trains an Escher agent.

  Args:
      cfg: (TrainConfig) The configuration for the training.
      agent: (Agent) The Escher agent to be trained.
  """
  device = torch.device(cfg.device)
  agent.avg_policy_net.to(device)
  for i in range(len(agent.regret_nets)):
    agent.regret_nets[i].to(device)
    agent.value_nets[i].to(device)

  for _ in range(cfg.iterations):
    _train_regret(cfg, agent)

    if agent.t % cfg.evaluation_interval == 0:
      _train_avg_policy(agent)
      if cfg.nashconv:
        conv = _calc_nashconv(cfg.game, agent)
        logging.info(
            "iteration %d states %d nashconv %f",
            agent.t,
            agent.num_touched,
            conv,
        )
      reward = _play_against_random(cfg.game, agent, cfg.games_vs_random)
      logging.info(
          "iteration %d states %d reward_vs_random %f",
          agent.t,
          agent.num_touched,
          reward,
      )

    agent.t += 1


def _play_once_against_random(game, agent):
  """Plays one game against a random policy and returns the reward."""
  reward = 0
  for player in range(game.num_players()):
    state = game.new_initial_state()
    while not state.is_terminal():
      if state.is_chance_node():
        outcomes, probs = zip(*state.chance_outcomes())
        a = np.random.choice(outcomes, p=probs)
        state.apply_action(a)
        continue

      if state.current_player() == player:
        policy = agent.action_probabilities(state)
      else:
        mask = np.array(state.legal_actions_mask(), dtype=int)
        policy = mask / np.sum(mask)
      action = np.random.choice(range(len(policy)), p=policy)
      state.apply_action(action)

    reward += state.returns()[player]

  return reward / game.num_players()


def _play_against_random(game, agent, n):
  reward = 0
  for _ in range(n):
    reward += _play_once_against_random(game, agent)
  return reward / n


def _calc_nashconv(game, agent):
  """Calculates the NashConv of the current policy."""
  def _action_probabilities(state):
    probs = agent.action_probabilities(state)

    prob_dict = {}
    for a, m in enumerate(state.legal_actions_mask()):
      if m == 1:
        prob_dict[a] = probs[a]
    return prob_dict

  policy = open_spiel.python.policy.tabular_policy_from_callable(
      game, _action_probabilities
  )
  conv = open_spiel.python.algorithms.exploitability.nash_conv(game, policy)
  return conv


class Transition(typing.NamedTuple):
  history: np.ndarray
  importance: np.ndarray
  action: np.ndarray
  returns: np.ndarray


class StateActionValue(typing.NamedTuple):
  state: np.ndarray
  action: np.ndarray
  value: np.ndarray


class StateRegret(typing.NamedTuple):
  state: np.ndarray
  regret: np.ndarray
  mask: np.ndarray
  t: np.ndarray


class Behaviour(typing.NamedTuple):
  state: np.ndarray
  policy: np.ndarray
  t: np.ndarray


def _state_history(num_players, state):
  history = []
  for p in range(num_players):
    history += state.information_state_tensor(p)
  return np.array(history, dtype=float)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utilities for AlphaZero."""

from typing import Any, Callable

import chex
from flax import linen
from flax import nnx
import jax
import jax.numpy as jnp

AVIALABLE_APIS = ["nnx", "linen"]

# pylint: disable=g-bare-generic
# pylint: disable=g-import-not-at-top
# pylint: disable=line-too-long


def api_selector(api_version):
  """Allows to choose an implementation API.

  Args:
    api_version (str): either of {AVIALABLE_APIS}

  Returns:
    model_lib: the model library with the chosen API.

  Raises:
    ValueError: if the user used smth different
  """
  if api_version == "nnx":
    from open_spiel.python.algorithms.alpha_zero import model_nnx as model_lib
  elif api_version == "linen":
    from open_spiel.python.algorithms.alpha_zero import model_linen as model_lib
  else:
    raise ValueError(f"Only {AVIALABLE_APIS} APIs are implmented")

  return model_lib


# These should be importable modes for inference only.
# I.e., you should call the inference using a saved checkpoint to use
# the APIs interchangeably
# See examples in `model_test.py`
def linen_to_nnx(model: linen.Module, seed: int = 0) -> nnx.bridge.ToNNX:
  # NOTE: could be issues with handling_randomness
  model = nnx.bridge.ToNNX(model, rngs=nnx.Rngs(seed))
  return model


def nnx_to_linen(
    model_class: nnx.Module,
    sample_shape: tuple[int, ...],
    seed: int = 0,
    *args: Any,
    **kwargs: Any,
) -> tuple[Callable, nnx.bridge.ToLinen]:
  new_model = nnx.bridge.ToLinen(model_class, *args, **kwargs)
  variables = new_model.init(
      jax.random.key(seed), (1, *sample_shape), train=False
  )
  return new_model.apply, variables


def flatten(x: chex.Array) -> chex.Array:
  """Flattens the array, i.e. reshapes it in a 1D shape."""
  return x.reshape(-1)


@chex.dataclass(frozen=True)
class TrainInput:
  """Inputs of the model: o_t, mask_t, π(a_t|o_t), v_t."""

  observation: chex.Array
  legals_mask: chex.Array
  policy: chex.Array
  value: chex.Array

  @staticmethod
  def stack(train_inputs: list["TrainInput"]) -> "TrainInput":
    return jax.tree.map(lambda *x: jnp.stack(x), *train_inputs)


@chex.dataclass(frozen=True)
class Losses:
  """Losses: policy, value, L2."""

  policy: chex.Array
  value: chex.Array
  l2: chex.Array

  @property
  def total(self) -> float:
    return self.policy + self.value + self.l2

  def __str__(self) -> str:
    return (
        f"Losses(total: {self.total:.3f}, policy: {self.policy:.3f}, value:"
        f" {self.value:.3f}, l2: {self.l2:.3f})"
    )

  def __add__(self, other) -> "Losses":
    return Losses(
        policy=self.policy + other.policy,
        value=self.value + other.value,
        l2=self.l2 + other.l2,
    )

  def __truediv__(self, n) -> "Losses":
    return Losses(policy=self.policy / n, value=self.value / n, l2=self.l2 / n)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest

from open_spiel.python.algorithms import minimax_solver
import pyspiel


class MinimaxSolverTest(absltest.TestCase):

  def test_compute_game_optimal_value(self):
    tic_tac_toe = pyspiel.load_game("tic_tac_toe")
    solver = minimax_solver.MinimaxSolver("tic_tac_toe")
    solver.solve()
    state = tic_tac_toe.new_initial_state()
    action_values = solver.action_values_from_state(state)
    print(f"Optimal action values at initial state: {action_values}")
    self.assertAlmostEqual(action_values.max(), 0.0)

  def test_compute_game_trembling_hand_optimal_value(self):
    tic_tac_toe = pyspiel.load_game("tic_tac_toe")
    solver = minimax_solver.MinimaxSolver("tic_tac_toe", epsilon=0.1)
    solver.solve()
    state = tic_tac_toe.new_initial_state()
    action_values = solver.action_values_from_state(state)
    print(f"Trembling opt. action values at initial state: {action_values}")
    # center is best in this case
    self.assertGreater(action_values.max(), 0.0)
    self.assertEqual(action_values.max(), action_values[4])
    # sides all equal, corners all equal, and corners better than sides
    self.assertAlmostEqual(action_values[0], action_values[2])
    self.assertAlmostEqual(action_values[2], action_values[6])
    self.assertAlmostEqual(action_values[6], action_values[8])
    self.assertAlmostEqual(action_values[1], action_values[3])
    self.assertAlmostEqual(action_values[3], action_values[5])
    self.assertAlmostEqual(action_values[5], action_values[7])
    self.assertGreater(action_values[0], action_values[1])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Ant Foraging on a Grid - A multi-agent foraging simulation game.

This game simulates ants foraging for food on a grid. Multiple ants start from
a nest and must navigate to find food sources, then return the food to the
nest. The game incorporates pheromone trails that ants can leave and follow.

Key mechanics:
- Grid-based world with nest(s), food source(s), and obstacles
- Ants can move in 4 directions (up, down, left, right) or stay
- Pheromones decay over time and influence ant movement heuristics
- Cooperative: all ants work together to maximize food collected
- Terminal condition: fixed number of turns or all food collected

This is a good testbed for multi-agent coordination and swarm intelligence
algorithms.
"""

# pylint: disable=protected-access


import enum

import numpy as np

from open_spiel.python.observation import IIGObserverForPublicInfoGame
import pyspiel


class CellType(enum.IntEnum):
  """Types of cells on the grid."""

  EMPTY = 0
  NEST = 1
  FOOD = 2
  OBSTACLE = 3


class Action(enum.IntEnum):
  """Possible actions for an ant."""

  STAY = 0
  UP = 1
  DOWN = 2
  LEFT = 3
  RIGHT = 4


# Movement deltas for each action (row, col)
_ACTION_DELTAS = {
    Action.STAY: (0, 0),
    Action.UP: (-1, 0),
    Action.DOWN: (1, 0),
    Action.LEFT: (0, -1),
    Action.RIGHT: (0, 1),
}

# Default game parameters
_DEFAULT_PARAMS = {
    "grid_size": 8,
    "num_ants": 2,
    "num_food": 3,
    "max_turns": 50,
    "pheromone_decay": 0.9,
}

_NUM_ACTIONS = len(Action)


def _make_game_type(params):
  """Create the game type with the given parameters."""
  num_ants = params.get("num_ants", _DEFAULT_PARAMS["num_ants"])
  return pyspiel.GameType(
      short_name="python_ant_foraging",
      long_name="Python Ant Foraging",
      dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
      chance_mode=pyspiel.GameType.ChanceMode.DETERMINISTIC,
      information=pyspiel.GameType.Information.PERFECT_INFORMATION,
      utility=pyspiel.GameType.Utility.IDENTICAL,
      reward_model=pyspiel.GameType.RewardModel.TERMINAL,
      max_num_players=num_ants,
      min_num_players=num_ants,
      provides_information_state_string=True,
      provides_information_state_tensor=False,
      provides_observation_string=True,
      provides_observation_tensor=True,
      parameter_specification={},
  )


# Default game type for registration
_GAME_TYPE = _make_game_type(_DEFAULT_PARAMS)


def _make_game_info(params):
  """Create game info based on parameters."""
  num_ants = params.get("num_ants", _DEFAULT_PARAMS["num_ants"])
  num_food = params.get("num_food", _DEFAULT_PARAMS["num_food"])
  max_turns = params.get("max_turns", _DEFAULT_PARAMS["max_turns"])
  return pyspiel.GameInfo(
      num_distinct_actions=_NUM_ACTIONS,
      max_chance_outcomes=0,
      num_players=num_ants,
      min_utility=0.0,
      max_utility=float(num_food),
      utility_sum=None,  # Not constant sum
      max_game_length=max_turns * num_ants,
  )


class AntForagingGame(pyspiel.Game):
  """The Ant Foraging game."""

  def __init__(self, params=None):
    """Initialize the game with given parameters."""
    params = params or {}
    self._grid_size = params.get("grid_size", _DEFAULT_PARAMS["grid_size"])
    self._num_ants = params.get("num_ants", _DEFAULT_PARAMS["num_ants"])
    self._num_food = params.get("num_food", _DEFAULT_PARAMS["num_food"])
    self._max_turns = params.get("max_turns", _DEFAULT_PARAMS["max_turns"])
    self._pheromone_decay = params.get(
        "pheromone_decay", _DEFAULT_PARAMS["pheromone_decay"]
    )

    game_type = _make_game_type(params)
    game_info = _make_game_info(params)
    super().__init__(game_type, game_info, params)

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return AntForagingState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an observer object for the game."""
    if (iig_obs_type is None) or (
        iig_obs_type.public_info and not iig_obs_type.perfect_recall
    ):
      return AntForagingObserver(self._grid_size, self._num_ants, params)
    return IIGObserverForPublicInfoGame(iig_obs_type, params)

  @property
  def grid_size(self):
    return self._grid_size

  @property
  def num_ants(self):
    return self._num_ants

  @property
  def num_food(self):
    return self._num_food

  @property
  def max_turns(self):
    return self._max_turns

  @property
  def pheromone_decay(self):
    return self._pheromone_decay


class AntForagingState(pyspiel.State):
  """State of the Ant Foraging game."""

  def __init__(self, game):
    """Constructor - sets up the initial game state."""
    super().__init__(game)
    self._game = game
    self._grid_size = game.grid_size
    self._num_ants = game.num_ants
    self._num_food = game.num_food
    self._max_turns = game.max_turns
    self._pheromone_decay = game.pheromone_decay

    # Initialize the grid
    self._grid = np.zeros((self._grid_size, self._grid_size), dtype=np.int32)

    # Place nest in center
    nest_pos = self._grid_size // 2
    self._nest_pos = (nest_pos, nest_pos)
    self._grid[self._nest_pos] = CellType.NEST

    # Place food sources randomly (avoiding nest and edges)
    self._food_positions = []
    np.random.seed(42)  # Deterministic for reproducibility
    attempts = 0
    while len(self._food_positions) < self._num_food and attempts < 100:
      row = np.random.randint(1, self._grid_size - 1)
      col = np.random.randint(1, self._grid_size - 1)
      if (row, col) != self._nest_pos and (
          row,
          col,
      ) not in self._food_positions:
        self._food_positions.append((row, col))
        self._grid[row, col] = CellType.FOOD
      attempts += 1

    # Initialize ant positions (all start at nest)
    self._ant_positions = [self._nest_pos for _ in range(self._num_ants)]

    # Track which ants are carrying food
    self._carrying_food = [False] * self._num_ants

    # Pheromone grids (one for "to food" and one for "to nest")
    self._pheromone_to_food = np.zeros(
        (self._grid_size, self._grid_size), dtype=np.float32
    )
    self._pheromone_to_nest = np.zeros(
        (self._grid_size, self._grid_size), dtype=np.float32
    )

    # Game state tracking
    self._current_player = 0
    self._turn_count = 0
    self._food_collected = 0
    self._is_terminal = False

  def current_player(self):
    """Returns the current player, or TERMINAL if the game is over."""
    if self._is_terminal:
      return pyspiel.PlayerId.TERMINAL
    return self._current_player

  def _legal_actions(self, player):
    """Returns the list of legal actions for a player."""
    if player != self._current_player:
      return []

    actions = [Action.STAY]
    row, col = self._ant_positions[player]

    # Check each direction
    for action in [Action.UP, Action.DOWN, Action.LEFT, Action.RIGHT]:
      dr, dc = _ACTION_DELTAS[action]
      new_row, new_col = row + dr, col + dc
      if self._is_valid_position(new_row, new_col):
        actions.append(action)

    return sorted(actions)

  def _is_valid_position(self, row, col):
    """Check if a position is within bounds and not an obstacle."""
    if row < 0 or row >= self._grid_size:
      return False
    if col < 0 or col >= self._grid_size:
      return False
    if self._grid[row, col] == CellType.OBSTACLE:
      return False
    return True

  def _apply_action(self, action):
    """Applies the action for the current player."""
    ant_idx = self._current_player
    row, col = self._ant_positions[ant_idx]

    # Move the ant
    dr, dc = _ACTION_DELTAS[action]
    new_row, new_col = row + dr, col + dc

    if self._is_valid_position(new_row, new_col):
      self._ant_positions[ant_idx] = (new_row, new_col)
      row, col = new_row, new_col

    # Check for food pickup
    if not self._carrying_food[ant_idx] and (row, col) in self._food_positions:
      self._carrying_food[ant_idx] = True
      self._food_positions.remove((row, col))
      self._grid[row, col] = CellType.EMPTY
      # Leave pheromone trail indicating food was here
      self._pheromone_to_food[row, col] = 1.0

    # Check for food delivery at nest
    if self._carrying_food[ant_idx] and (row, col) == self._nest_pos:
      self._carrying_food[ant_idx] = False
      self._food_collected += 1
      # Leave pheromone trail to nest
      self._pheromone_to_nest[row, col] = 1.0

    # Update pheromone based on ant's state
    if self._carrying_food[ant_idx]:
      # Ant carrying food leaves "to nest" pheromone
      self._pheromone_to_nest[row, col] = min(
          1.0, self._pheromone_to_nest[row, col] + 0.3
      )
    else:
      # Ant searching leaves "to food" pheromone when near food
      if self._pheromone_to_food[row, col] > 0:
        self._pheromone_to_food[row, col] = min(
            1.0, self._pheromone_to_food[row, col] + 0.1
        )

    # Move to next player
    self._current_player = (self._current_player + 1) % self._num_ants

    # If we've cycled through all ants, that's one turn
    if self._current_player == 0:
      self._turn_count += 1
      # Decay pheromones at end of each round
      self._pheromone_to_food *= self._pheromone_decay
      self._pheromone_to_nest *= self._pheromone_decay

    # Check terminal conditions
    if self._turn_count >= self._max_turns:
      self._is_terminal = True
    if self._food_collected >= self._num_food:
      self._is_terminal = True

  def _action_to_string(self, player, action):
    """Converts an action to a human-readable string."""
    action_names = {
        Action.STAY: "stay",
        Action.UP: "up",
        Action.DOWN: "down",
        Action.LEFT: "left",
        Action.RIGHT: "right",
    }
    return f"ant{player}:{action_names.get(action, '?')}"

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._is_terminal

  def returns(self):
    """Returns the total reward for each player."""
    # Cooperative game - all players share the same score
    score = float(self._food_collected)
    return [score] * self._num_ants

  def __str__(self):
    """Returns a string representation of the current state."""
    lines = []
    lines.append(
        f"Turn {self._turn_count}/{self._max_turns}, "
        f"Food: {self._food_collected}/{self._num_food}"
    )

    # Build grid visualization
    grid_chars = []
    for row in range(self._grid_size):
      row_chars = []
      for col in range(self._grid_size):
        # Check for ants first
        ant_here = None
        for i, pos in enumerate(self._ant_positions):
          if pos == (row, col):
            ant_here = i
            break

        if ant_here is not None:
          if self._carrying_food[ant_here]:
            row_chars.append(str(ant_here).upper())
          else:
            row_chars.append(str(ant_here))
        elif self._grid[row, col] == CellType.NEST:
          row_chars.append("N")
        elif self._grid[row, col] == CellType.FOOD:
          row_chars.append("F")
        elif self._grid[row, col] == CellType.OBSTACLE:
          row_chars.append("#")
        else:
          # Show pheromone intensity
          pf = self._pheromone_to_food[row, col]
          pn = self._pheromone_to_nest[row, col]
          if pf > 0.5 or pn > 0.5:
            row_chars.append("~")
          elif pf > 0.1 or pn > 0.1:
            row_chars.append(".")
          else:
            row_chars.append(" ")
      grid_chars.append("".join(row_chars))

    # Add border
    border = "+" + "-" * self._grid_size + "+"
    lines.append(border)
    for row_str in grid_chars:
      lines.append("|" + row_str + "|")
    lines.append(border)

    # Add legend
    lines.append("Legend: N=nest, F=food, 0-9=ant, A-Z=ant+food")

    return "\n".join(lines)


class AntForagingObserver:
  """Observer for the Ant Foraging game."""

  def __init__(self, grid_size, num_ants, params=None):
    """Initialize the observer."""
    if params:
      raise ValueError(f"Observation parameters not supported; got {params}")

    self._grid_size = grid_size
    self._num_ants = num_ants

    # Observation tensor layout:
    # - Grid channels: empty, nest, food, obstacle (4)
    # - Ant positions: one channel per ant (num_ants)
    # - Carrying food: one channel per ant (num_ants)
    # - Pheromone channels: to_food, to_nest (2)
    num_channels = 4 + num_ants * 2 + 2
    shape = (num_channels, grid_size, grid_size)
    self.tensor = np.zeros(np.prod(shape), np.float32)
    self.dict = {"observation": np.reshape(self.tensor, shape)}

  def set_from(self, state, player):
    """Updates the observation tensor from the given state."""
    del player  # Observation is the same for all players
    obs = self.dict["observation"]
    obs.fill(0)

    # Grid cell types
    for row in range(self._grid_size):
      for col in range(self._grid_size):
        cell = state._grid[row, col]
        obs[cell, row, col] = 1.0

    # Ant positions
    for i, (row, col) in enumerate(state._ant_positions):
      obs[4 + i, row, col] = 1.0

    # Carrying food
    for i, carrying in enumerate(state._carrying_food):
      if carrying:
        row, col = state._ant_positions[i]
        obs[4 + self._num_ants + i, row, col] = 1.0

    # Pheromones
    obs[-2] = state._pheromone_to_food
    obs[-1] = state._pheromone_to_nest

  def string_from(self, state, player):
    """Returns string observation."""
    del player
    return str(state)


# Register the game with OpenSpiel
pyspiel.register_game(_GAME_TYPE, AntForagingGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An AlphaZero model with a policy and value heads using flax.linen."""

import functools
import os
from typing import Any, Callable, Optional, Sequence
import warnings

import chex
import flax
from flax import training
from flax import traverse_util
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
import optax
import orbax
import orbax.checkpoint

from open_spiel.python.algorithms.alpha_zero import utils

# pylint: disable=g-bare-generic
# pylint: disable=redefined-outer-name

flax.config.update("flax_use_orbax_checkpointing", True)
warnings.warn("Pay attention that you've been using the `linen` api")

activations_dict = {
    "celu": nn.celu,
    "elu": nn.elu,
    "gelu": nn.gelu,
    "glu": nn.glu,
    "hard_sigmoid": nn.hard_sigmoid,
    "hard_silu": nn.hard_silu,  # Alias for hard_swish
    "hard_swish": nn.hard_swish,  # Alias for hard_silu
    "hard_tanh": nn.hard_tanh,
    "leaky_relu": nn.leaky_relu,
    "log_sigmoid": nn.log_sigmoid,
    "log_softmax": nn.log_softmax,
    "relu": nn.relu,
    "selu": nn.selu,
    "sigmoid": nn.sigmoid,
    "silu": nn.silu,
    "soft_sign": nn.soft_sign,
    "softmax": nn.softmax,
    "softplus": nn.softplus,
    "swish": nn.swish,
    "tanh": nn.tanh,
}


class Activation(nn.Module):
  """A simple `nn.Module` class wrapper for the activations."""

  activation_name: str

  @nn.compact
  def __call__(self, x: chex.Array) -> chex.Array:
    # if the activation not found or None just passes the identity function
    return activations_dict.get(self.activation_name, lambda x: x)(x)


class MLPBlock(nn.Module):
  features: int
  activation: str = "relu"

  @nn.compact
  def __call__(self, x: chex.Array) -> chex.Array:
    y = nn.Dense(features=self.features)(x)
    y = Activation(self.activation)(y)
    return y


class ConvBlock(nn.Module):
  """A convolution block for the AlphaZero model."""

  features: int
  kernel_size: tuple[int, int]
  activation: Optional[str] = "relu"

  @nn.compact
  def __call__(self, x: chex.Array, training: bool = False) -> chex.Array:
    y = nn.Conv(
        features=self.features, kernel_size=self.kernel_size, padding="SAME"
    )(x)
    y = nn.BatchNorm(use_running_average=not training, axis_name="batch")(y)
    y = Activation(self.activation)(y)
    return y


class ResidualBlock(nn.Module):
  filters: int
  kernel_size: tuple[int, int]
  activation: Optional[str] = "relu"

  @nn.compact
  def __call__(self, x: chex.Array, training: bool = False) -> chex.Array:
    residual = x
    y = ConvBlock(self.filters, self.kernel_size, self.activation)(x, training)
    y = ConvBlock(self.filters, self.kernel_size, None)(y, training)
    y = Activation(self.activation)(y + residual)
    return y


class PolicyHead(nn.Module):
  """A policy head for the AlphaZero model."""

  model_type: str
  nn_width: int
  output_size: int
  activation: Optional[str] = "relu"

  @nn.compact
  def __call__(self, x: chex.Array, training: bool = False) -> chex.Array:
    if self.model_type != "mlp":
      x = ConvBlock(features=2, kernel_size=(1, 1), activation=self.activation)(
          x, training
      )
      x = utils.flatten(x)

    x = MLPBlock(self.nn_width, activation=self.activation)(x)
    policy_logits = MLPBlock(self.output_size, None)(x)
    return policy_logits


class ValueHead(nn.Module):
  """A value head for the AlphaZero model."""
  model_type: str
  nn_width: int
  activation: Optional[str] = "relu"

  @nn.compact
  def __call__(self, x: chex.Array, training: bool = False) -> chex.Array:
    if self.model_type != "mlp":
      x = ConvBlock(features=1, kernel_size=(1, 1), activation=self.activation)(
          x, training
      )
      x = utils.flatten(x)

    x = MLPBlock(self.nn_width, self.activation)(x)
    values = MLPBlock(1, "tanh")(x)
    return values


class AlphaZeroModel(nn.Module):
  """An AlphaZero style model with a policy and value head.

  This supports three types of models: mlp, conv2d and resnet.

  All models have a shared torso stack with two output heads: policy and value.
  They have same meaning as in the AlphaGo Zero and AlphaZero papers. The resnet
  model copies the one in that paper when set with width 256 and depth 20. The
  conv2d model is the same as the resnet except uses a conv+batchnorm+relu
  instead of the res blocks. The mlp model uses dense layers instead of conv,
  and drops batch norm.

  Links to relevant articles/papers:
    https://deepmind.com/blog/article/alphago-zero-starting-scratch has an open
      access link to the AlphaGo Zero nature paper.
    https://deepmind.com/blog/article/alphazero-shedding-new-light-grand-games-chess-shogi-and-go
      has an open access link to the AlphaZero science paper.

  All are parameterized by their input (observation) shape and output size
  (number of actions), though the conv2d and resnet might only work with games
  that have spatial data (ie 3 non-batch dimensions, eg: connect four would
  work, but not poker).

  The depth is the number of blocks in the torso, where the definition of a
  block varies by model. For a resnet it's a resblock which is two conv2ds,
  batch norms and relus, and an addition. For conv2d it's a conv2d, a batch norm
  and a relu. For mlp it's a dense plus relu.

  The width is the number of filters for any conv2d and the number of hidden
  units for any dense layer.
  """

  model_type: str
  input_shape: chex.Shape
  output_size: int
  nn_width: int
  nn_depth: int
  activation: Optional[str] = "relu"

  @nn.compact
  def __call__(
      self, observations: chex.Array, training: bool
  ) -> tuple[chex.Array, chex.Array]:
    # torso:
    x = observations
    if self.model_type == "mlp":
      x = utils.flatten(observations)
      for _ in range(self.nn_depth):
        x = MLPBlock(features=self.nn_width, activation=self.activation)(x)
    elif self.model_type == "conv2d":
      x = observations.reshape(self.input_shape)
      for _ in range(self.nn_depth):
        x = ConvBlock(
            features=self.nn_width,
            kernel_size=(3, 3),
            activation=self.activation,
        )(x, training)
    elif self.model_type == "resnet":
      x = observations.reshape(self.input_shape)
      x = ConvBlock(
          features=self.nn_width, kernel_size=(3, 3), activation=self.activation
      )(x, training)
      for _ in range(self.nn_depth):
        x = ResidualBlock(
            filters=self.nn_width,
            kernel_size=(3, 3),
            activation=self.activation,
        )(x, training)
    else:
      raise ValueError(f"Unknown model type: {self.model_type}")

    policy_logits = PolicyHead(
        model_type=self.model_type,
        nn_width=self.nn_width,
        output_size=self.output_size,
    )(x, training)
    value_out = ValueHead(model_type=self.model_type, nn_width=self.nn_width)(
        x, training
    )

    return policy_logits, value_out.squeeze(-1)


class TrainState(training.train_state.TrainState):
  batch_stats: Any


class Model:
  """Basic model class."""
  valid_model_types = ["mlp", "conv2d", "resnet"]

  def __init__(
      self,
      state: TrainState,
      path: str,
      update_step_fn: Callable,
  ) -> None:
    self._state = state
    self._path = path
    self._update_step_fn = update_step_fn

    self._checkpointer = None
    if self._path is not None:
      self._checkpointer = orbax.checkpoint.PyTreeCheckpointer()

  @classmethod
  def _create_train_state(
      cls,
      apply_fn: Callable,
      variables: flax.core.FrozenDict,
      optimiser: optax.GradientTransformation,
  ) -> TrainState:
    return TrainState.create(
        apply_fn=apply_fn,
        params=variables["params"],
        tx=optimiser,
        batch_stats=variables.get("batch_stats", {}),
    )

  @classmethod
  def build_model(
      cls,
      model_type: str,
      input_shape: chex.Shape,
      output_size: chex.Numeric,
      nn_width: int,
      nn_depth: int,
      weight_decay: float,
      learning_rate: float,
      path: str,
      seed: int = 0,
      decouple_weight_decay: bool = False,
  ) -> "Model":
    """Builds the model."""

    if model_type not in cls.valid_model_types:
      raise ValueError(
          f"Invalid model type: {model_type}, expected one of:"
          f" {cls.valid_model_types}"
      )

    if isinstance(input_shape, int):
      input_shape = (input_shape,)
    elif isinstance(input_shape, list):
      input_shape = tuple(input_shape)

    while len(input_shape) < 3:
      input_shape = input_shape + (1,)

    model = AlphaZeroModel(
        model_type=model_type,
        input_shape=input_shape,
        output_size=output_size,
        nn_width=nn_width,
        nn_depth=nn_depth,
    )

    def mask_only_biases(params):
      flat_params = traverse_util.flatten_dict(params)
      flat_mask = {
          path: not ((path[-1] == "bias") or ("BatchNorm_0" in path))
          for path in flat_params.keys()
      }
      return traverse_util.unflatten_dict(flat_mask)

    rng = jax.random.PRNGKey(seed)
    variables = model.init(rng, jnp.ones(input_shape), False)

    if not decouple_weight_decay:
      optimiser = optax.adam(learning_rate=learning_rate)
    else:
      optimiser = optax.adamw(
          learning_rate=learning_rate,
          weight_decay=weight_decay,
          mask=mask_only_biases,
      )

    state = cls._create_train_state(model.apply, variables, optimiser)

    @jax.jit
    def update_step_fn(
        state, observations, legals_mask, policy_targets, value_targets
    ):
      def loss_fn(
          params,
          batch_stats,
          observations,
          legals_mask,
          policy_targets,
          value_targets,
      ):
        variables = {"params": params, "batch_stats": batch_stats}

        @functools.partial(
            jax.vmap,
            in_axes=(0, 0, 0, 0),
            out_axes=(0, None),
            axis_name="batch",
        )
        def _per_example_loss(
            observations, legals_mask, policy_targets, value_targets
        ):
          (policy_logits, value_preds), new_model_state = state.apply_fn(
              variables, observations, training=True, mutable=["batch_stats"]
          )

          policy_logits = jnp.where(
              legals_mask,
              policy_logits,
              jnp.full_like(policy_logits, jnp.finfo(jnp.float32).min),
          )
          policy_loss = optax.softmax_cross_entropy(
              policy_logits, policy_targets
          )
          value_loss = optax.l2_loss(value_preds, value_targets)

          return (policy_loss, value_loss), new_model_state["batch_stats"]

        (policy_loss, value_loss), new_model_state = _per_example_loss(
            observations, legals_mask, policy_targets, value_targets
        )
        policy_loss = policy_loss.mean()
        value_loss = value_loss.mean()

        l2_reg_loss = (
            optax.tree_utils.tree_norm(
                jax.tree.map(
                    lambda p, m: p * m, params, mask_only_biases(params)
                ),
                ord=2,
                squared=True,
            )
            * weight_decay
        )

        total_loss = policy_loss + value_loss
        if not decouple_weight_decay:
          total_loss = total_loss + l2_reg_loss

        return total_loss, (
            policy_loss,
            value_loss,
            l2_reg_loss,
            new_model_state,
        )

      grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
      (_, (policy_loss, value_loss, l2_reg_loss, new_batch_stats)), grads = (
          grad_fn(
              state.params,
              state.batch_stats,
              observations,
              legals_mask,
              policy_targets,
              value_targets,
          )
      )

      new_state = state.apply_gradients(
          grads=grads, batch_stats=new_batch_stats
      )
      return new_state, (policy_loss, value_loss, l2_reg_loss)

    return cls(state, path, update_step_fn)

  @classmethod
  def from_checkpoint(cls, checkpoint_path, model_args=None):
    if model_args is None:
      raise ValueError("model_args must be provided for checkpoint loading")

    model = cls.build_model(**model_args)
    model._state = cls.load_checkpoint(checkpoint_path)

    return model

  @property
  def num_trainable_variables(self) -> int:
    return sum(np.prod(p.shape) for p in jax.tree.leaves(self._state.params))

  def print_trainable_variables(self) -> None:
    flat_params, _ = jax.tree.flatten(self._state.params)
    for i, p in enumerate(flat_params):
      print(f"Param {i}: {p.shape}")

  def inference(
      self, observation: list, legals_mask: list
  ) -> tuple[chex.Array, chex.Array]:
    """Returns the value and policy for the given state."""
    observation = jnp.asarray(observation, dtype=jnp.float32)
    legals_mask = jnp.asarray(legals_mask, dtype=jnp.bool)

    # @jax.jit  # --- with it the model is much slower than without...
    def _predict(params, batch_stats, observation: chex.Array):
      policy_logits, value = self._state.apply_fn(
          {"params": params, "batch_stats": batch_stats},
          observation,
          training=False,
          mutable=False,
      )
      policy_logits = jnp.where(
          legals_mask,
          policy_logits,
          jnp.full_like(policy_logits, jnp.finfo(jnp.float32).min),
      )
      policy = nn.softmax(policy_logits, axis=-1)
      return value, policy

    value, policy = _predict(
        self._state.params, self._state.batch_stats, observation
    )

    return value, policy

  def update(self, batch: Sequence[utils.TrainInput]) -> utils.Losses:
    self._state, (policy_loss, value_loss, l2_reg_loss) = self._update_step_fn(
        self._state,
        batch.observation,
        batch.legals_mask,
        batch.policy,
        batch.value,
    )

    return utils.Losses(policy=policy_loss, value=value_loss, l2=l2_reg_loss)

  def save_checkpoint(self, step: int, device=None) -> int:
    """Saves a checkpoint of the model."""
    jax.block_until_ready(self._state)
    path = os.path.join(self._path, f"checkpoint-{step}")

    if device is None:
      device = jax.local_devices()[0]

    sharded_state = jax.tree_util.tree_map(
        lambda x: jax.device_put(x, jax.sharding.SingleDeviceSharding(device)),
        self._state,
    )
    if self._checkpointer:
      self._checkpointer.save(
          path,
          args=orbax.checkpoint.args.PyTreeSave(item=sharded_state),
          force=True,
      )

    return step

  def load_checkpoint(self, step: int | str, device: str = None) -> None:
    """Loads a checkpoint of the model."""
    target = self._state
    path = os.path.join(self._path, f"checkpoint-{step}")

    if device is None:
      device = jax.local_devices()[0]

    sharding = jax.sharding.SingleDeviceSharding(device)

    restore_args_tree = jax.tree.map(
        lambda _: orbax.checkpoint.type_handlers.ArrayRestoreArgs(
            sharding=sharding
        ),
        target,
    )

    if self._checkpointer:
      self._state = self._checkpointer.restore(
          path, item=orbax.checkpoint.args.PyTreeRestore(item=restore_args_tree)
      )
      jax.block_until_ready(self._state)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for tic_tac_toe."""

from typing import Any

from absl.testing import absltest
from absl.testing import parameterized

import pyspiel
ttt = pyspiel.tic_tac_toe


class GamesTicTacToeTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.game = pyspiel.load_game("tic_tac_toe")
    self.ref_state = self.game.new_initial_state()
    self.ref_state.apply_action(4)

  def test_constants(self):
    self.assertEqual(ttt.NUM_ROWS, 3)
    self.assertEqual(ttt.NUM_COLS, 3)
    self.assertEqual(ttt.NUM_CELLS, 9)
    self.assertEqual(ttt.CellState.EMPTY.value, 0)
    self.assertEqual(ttt.CellState.NOUGHT.value, 1)
    self.assertEqual(ttt.CellState.CROSS.value, 2)

  @parameterized.parameters(
      (0, ttt.CellState.CROSS),
      (1, ttt.CellState.NOUGHT),
  )
  def test_player_to_cellstate(
      self, player: int, expected_cellstate: ttt.CellState
  ):
    self.assertEqual(ttt.player_to_cellstate(player), expected_cellstate)

  @parameterized.parameters(
      (ttt.CellState.EMPTY, "."),
      (ttt.CellState.NOUGHT, "o"),
      (ttt.CellState.CROSS, "x"),
  )
  def test_cellstate_to_string(
      self, cellstate: ttt.CellState, expected_string: str
  ):
    self.assertEqual(ttt.cellstate_to_string(cellstate), expected_string)

  def test_board_at(self):
    self.assertEqual(self.ref_state.board_at(1, 1), ttt.CellState.CROSS)

  def test_board(self):
    state = self.game.new_initial_state()
    state.apply_action(0)
    state.apply_action(1)
    self.assertEqual(state.board(), [
        ttt.CellState.CROSS,
        ttt.CellState.NOUGHT] + [ttt.CellState.EMPTY] * 7)

  def test_json(self):
    state_struct = self.ref_state.to_struct()
    self.assertEqual(
        state_struct.board,
        [".", ".", ".", ".", "x", ".", ".", ".", "."],
    )
    self.assertEqual(state_struct.current_player, "o")
    json_from_struct = state_struct.to_json()
    state_json = self.ref_state.to_json()
    self.assertEqual(
        state_json,
        '{"board":[".",".",".",".","x",".",".",".","."],"current_player":"o"}',
    )
    self.assertEqual(json_from_struct, state_json)
    state_struct = ttt.TicTacToeStateStruct(state_json)
    self.assertEqual(state_struct.to_json(), state_json)

  @parameterized.parameters(
      dict(
          actions=[4],
          expected_dict={
              "board": [".", ".", ".", ".", "x", ".", ".", ".", "."],
              "current_player": "o",
          },
      ),
      dict(
          actions=[0, 1, 2],
          expected_dict={
              "board": ["x", "o", "x", ".", ".", ".", ".", ".", "."],
              "current_player": "o",
          },
      ),
      dict(
          actions=[0, 1, 3, 4, 6],  # X wins
          expected_dict={
              "board": ["x", "o", ".", "x", "o", ".", "x", ".", "."],
              "current_player": "Terminal",
          },
      ),
      dict(
          actions=[0, 3, 1, 4, 5, 6, 2],  # X wins
          expected_dict={
              "board": ["x", "x", "x", "o", "o", "x", "o", ".", "."],
              "current_player": "Terminal",
          },
      ),
  )
  def test_state_to_from_dict(
      self, actions: list[int], expected_dict: dict[str, Any]
  ):
    state = self.game.new_initial_state()
    for action in actions:
      state.apply_action(action)
      # Test round trip.
      new_state = self.game.new_initial_state(state.to_dict())
      self.assertEqual(new_state.to_dict(), state.to_dict())
    state_dict = state.to_dict()
    self.assertEqual(state_dict, expected_dict)

  @parameterized.parameters(
      dict(
          invalid_dict={
              "board": ["."],  # Wrong length.
              "current_player": "x",
          }
      ),
      dict(
          invalid_dict={
              "board": [".", ".", ".", ".", "x", ".", ".", ".", "."],
              "current_player": "x",  # Should be o's turn.
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "x", ".", ".", ".", ".", ".", ".", "."],
              "current_player": "o",  # Count difference > 1.
          }
      ),
      dict(
          invalid_dict={
              "board": ["o", ".", ".", ".", ".", ".", ".", ".", "."],
              "current_player": "o",  # Player o cannot start.
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "o", ".", ".", ".", ".", ".", ".", "."],
              "current_player": "o",  # Should be x's turn.
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "o", "y", ".", ".", ".", ".", ".", "."],
              "current_player": "x",  # Invalid character 'y'.
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "x", "x", "o", "o", ".", ".", ".", "."],
              "current_player": "o",  # Should be "Terminal".
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "x", "x", "o", "o", "o", ".", ".", "."],
              "current_player": "Terminal",  # Both players have a line.
          }
      ),
      dict(
          invalid_dict={
              "board": ["x", "x", "x", "o", "o", ".", "o", ".", "."],
              "current_player": "Terminal",  # x has a line but o just moved.
          }
      ),
  )
  def test_state_from_invalid_dict(self, invalid_dict: dict[str, Any]):
    with self.assertRaises(pyspiel.SpielError):
      self.game.new_initial_state(invalid_dict)

  def test_empty_starting_state(self):
    state = self.game.new_initial_state()
    state.apply_action(0)
    state.apply_action(1)
    state.apply_action(2)
    self.assertIsNone(state.starting_state())
    self.assertEmpty(state.starting_state_str())

  def test_starting_state_from_dict(self):
    state = self.game.new_initial_state(self.ref_state.to_dict())
    self.assertEmpty(state.history())
    self.assertEqual(str(state.starting_state()), str(self.ref_state))
    self.assertEqual(state.starting_state_str(), self.ref_state.to_json())

  def test_starting_state_from_json(self):
    state = self.game.new_initial_state(self.ref_state.to_json())
    self.assertEmpty(state.history())
    self.assertEqual(str(state.starting_state()), str(self.ref_state))

  def test_starting_state_unchanged_after_action(self):
    state = self.game.new_initial_state(self.ref_state.to_dict())
    state.apply_action(0)
    self.assertEqual(str(state.starting_state()), str(self.ref_state))

  def test_starting_state_unchanged_after_clone(self):
    state = self.game.new_initial_state(self.ref_state.to_dict())
    state.apply_action(0)
    clone = state.clone()
    self.assertEqual(str(clone.starting_state()), str(self.ref_state))

  def test_state_serialization(self):
    state = self.game.deserialize_state(self.ref_state.serialize())
    self.assertEqual(str(state), str(self.ref_state))

  def test_starting_state_serialization(self):
    state_with_starting = self.game.new_initial_state(self.ref_state.to_dict())
    state_after_serialization = self.game.deserialize_state(
        state_with_starting.serialize()
    )

    self.assertEqual(
        str(state_after_serialization.starting_state()), str(self.ref_state)
    )
    self.assertEmpty(state_after_serialization.history())

if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Deep CFR Algorithm.

See https://arxiv.org/abs/1811.00164.

The algorithm defines an `advantage` and `strategy` networks that compute
advantages used to do regret matching across information sets and to approximate
the strategy profiles of the game. To train these networks a reservoir buffer
(other data structures may be used) memory is used to accumulate samples to
train the networks.

This implementation uses skip connections as described in the paper if two
consecutive layers of the advantage or policy network have the same number
of units, except for the last connection. Before the last hidden layer
a layer normalization is applied.

"""

# pylint: disable=g-bare-generic

import collections
import functools
from typing import Callable, Iterable, NamedTuple

import chex
import flax.nnx as nn
import jax
import jax.numpy as jnp
import numpy as np
import optax

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
import pyspiel
# pylint: disable=g-importing-member
from open_spiel.python.utils.lru_cache import LRUCache


ILLEGAL_ACTION_LOGITS_PENALTY = jnp.finfo(jnp.float32).min


class AdvantageMemory(NamedTuple):
  """Advantage network memory buffer."""

  info_state: chex.Array
  iteration: chex.Array
  advantage: chex.Array
  legal_mask: chex.Array


class StrategyMemory(NamedTuple):
  """Strategy network memory buffer."""

  info_state: chex.Array
  iteration: chex.Array
  strategy_action_probs: chex.Array
  legal_mask: chex.Array


class ReservoirBufferState(NamedTuple):

  experience: chex.ArrayTree
  capacity: chex.Numeric
  add_calls: chex.Array
  is_full: chex.Array

  def __len__(self) -> int:
    return jnp.where(self.is_full, self.capacity, self.add_calls)


class ReservoirBuffer:
  """Allows uniform sampling over a stream of data.

  See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
  """

  @staticmethod
  @functools.partial(jax.jit, static_argnames=("capacity",))
  def init(
      capacity: chex.Numeric, experience: chex.ArrayTree
  ) -> ReservoirBufferState:
    """Initializes the reservoir buffer."""

    # Set experience value to be empty.
    experience = jax.tree.map(jnp.empty_like, experience)
    # Broadcast to [add_batch_size, ...]
    experience = jax.tree.map(
        lambda x: jnp.broadcast_to(x[jnp.newaxis, ...], (capacity, *x.shape)),
        experience,
    )
    return ReservoirBufferState(
        capacity=capacity,
        experience=experience,
        add_calls=jnp.array(0),
        is_full=jnp.array(False, dtype=jnp.bool),
    )

  @staticmethod
  @functools.partial(jax.jit, donate_argnums=(0,))
  def append(
      state: ReservoirBufferState, experience: chex.ArrayTree, rng: chex.PRNGKey
  ) -> ReservoirBufferState:
    """Potentially adds `experience` to the reservoir buffer.

    Args:
      state: `ReservoirBufferState`, current state of the buffer
      experience: data to be added to the reservoir buffer.
      rng: `chex.PRNGKey`, a random seed

    Returns:
      An updated `ReservoirBufferState`
    """

    # Note: count + 1 because the current item is the (count+1)-th item
    idx = jax.random.randint(rng, (), 0, state.add_calls + 1)

    # 2. Logic:
    # If buffer is not full, we always add at 'count'.
    # If buffer is full, we replace at 'idx' ONLY IF idx < capacity.
    is_full = state.is_full | (state.add_calls >= state.capacity)
    write_idx = jnp.where(is_full, idx, state.add_calls)
    should_update = write_idx < state.capacity

    def update_leaf(buffer_leaf, exp_leaf):
      new_val = jnp.where(should_update, exp_leaf, buffer_leaf[write_idx])
      return buffer_leaf.at[write_idx].set(new_val)

    new_experience = jax.tree.map(update_leaf, state.experience, experience)

    return ReservoirBufferState(
        capacity=state.capacity,
        experience=new_experience,
        add_calls=state.add_calls + 1,
        is_full=is_full,
    )

  @staticmethod
  @functools.partial(jax.jit, static_argnames="num_samples")
  def sample(rng: chex.PRNGKey, state: ReservoirBufferState, num_samples: int):
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      rng: `chex.PRNGKey`, a random state
      state: `ReservoirBufferState`, a buffer state
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      AssertionError: If there are less than `num_samples` elements in the
      buffer
    """
    max_size = jnp.where(state.is_full, state.capacity, state.add_calls)
    indices = jax.random.randint(
        rng, shape=(num_samples,), minval=0, maxval=max_size
    )
    return jax.tree.map(lambda x: x[indices], state.experience)


class MLP(nn.Module):
  """A simple MLP."""

  def __init__(
      self,
      input_size: int,
      hidden_sizes: Iterable[int],
      output_size: int,
      final_activation: Callable = lambda x: x,
      seed: int = 0,
  ) -> None:

    layers_ = []

    def _create_linear_block(in_features, out_features, act=nn.relu):
      return nn.Sequential(
          nn.Linear(
              in_features,
              out_features,
              kernel_init=nn.initializers.glorot_uniform(),
              rngs=nn.Rngs(seed),
          ),
          act,
      )

    # Input and Hidden layers
    for size in hidden_sizes:
      layers_.append(_create_linear_block(input_size, size, act=nn.relu))
      input_size = size
    # Output layer
    layers_.append(nn.LayerNorm(input_size, rngs=nn.Rngs(seed)))
    layers_.append(
        _create_linear_block(input_size, output_size, act=lambda x: x)
    )
    if final_activation:
      layers_.append(final_activation)
    self.model = nn.Sequential(*layers_)

  def __call__(self, x: chex.Array) -> chex.Array:
    outputs = self.model(x)
    return outputs

  # NOTE: reset is done a bit differently


@nn.vmap(in_axes=(None, 0), out_axes=0)
def forward(model, x):
  return model(x)


class DeepCFRSolver(policy.Policy):
  """Implements a solver for the Deep CFR Algorithm.

  See https://arxiv.org/abs/1811.00164.

  Define all networks and sampling buffers/memories.  Derive losses & learning
  steps. Initialize the game state and algorithmic variables.
  """

  def __init__(
      self,
      game,
      policy_network_layers=(256, 256),
      advantage_network_layers=(128, 128),
      num_iterations: int = 100,
      num_traversals: int = 20,
      learning_rate: float = 1e-4,
      batch_size_advantage: int = None,
      batch_size_strategy: int = None,
      memory_capacity: int = int(1e6),
      policy_network_train_steps: int = 5000,
      advantage_network_train_steps: int = 750,
      reinitialize_advantage_networks: bool = True,
      seed: int = 42,
      print_nash_convs: bool = False,
  ) -> None:
    """Initialize the Deep CFR algorithm.

    Args:
      game: Open Spiel game.
      policy_network_layers: (list[int]) Layer sizes of strategy net MLP.
      advantage_network_layers: (list[int]) Layer sizes of advantage net MLP.
      num_iterations: Number of iterations.
      num_traversals: Number of traversals per iteration.
      learning_rate: Learning rate.
      batch_size_advantage: (int) Batch size to sample from advantage memories.
      batch_size_strategy: (int) Batch size to sample from strategy memories.
      memory_capacity: Number of samples that can be stored in memory.
      policy_network_train_steps: Number of policy network training steps (one
        policy training iteration at the end).
      advantage_network_train_steps: Number of advantage network training steps
        (per iteration).
      reinitialize_advantage_networks: Whether to re-initialize the advantage
        network before training on each iteration.
            seed: (int) A random seed
      print_nash_convs: (bool) print explotability for each iteration, defaults
        to False
    """
    all_players = list(range(game.num_players()))
    super().__init__(game, all_players)

    self._game = game
    if game.get_type().dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS:
      # `_traverse_game_tree` does not take into account this option.
      raise ValueError("Simulatenous games are not supported.")
    self._batch_size_advantage = batch_size_advantage
    self._batch_size_strategy = batch_size_strategy
    self._policy_network_train_steps = policy_network_train_steps
    self._advantage_network_train_steps = advantage_network_train_steps
    self._policy_network_layers = policy_network_layers
    self._advantage_network_layers = advantage_network_layers
    self._num_players = game.num_players()
    self._root_node = self._game.new_initial_state()
    self._embedding_size = len(self._root_node.information_state_tensor(0))
    self._num_iterations = num_iterations
    self._num_traversals = num_traversals
    self._reinitialize_advantage_networks = reinitialize_advantage_networks
    self._num_actions = game.num_distinct_actions()
    self._iteration = 1
    self._learning_rate = learning_rate
    self._rngkey = jax.random.key(seed)
    self._print_nash_convs = print_nash_convs
    self._memory_capacity = int(memory_capacity)

    self._backend = jax.default_backend()

    # Initialize networks and memory buffers
    self._advantage_memories = [None] * self._num_players
    self._advantage_networks = [
        MLP(
            self._embedding_size,
            list(advantage_network_layers),
            self._num_actions,
            lambda x: x,
            seed + p,
        )
        for p in range(self._num_players)
    ]
    self._advantage_opt = [
        nn.Optimizer(
            self._advantage_networks[p],
            optax.adam(self._learning_rate),
            wrt=nn.Param,
        )
        for p in range(self._num_players)
    ]
    self._empty_advantage_states = [
        nn.state((self._advantage_networks[p], self._advantage_opt[p]))
        for p in range(self._num_players)
    ]
    self._advantage_graphdefs = [
        nn.graphdef((self._advantage_networks[p], self._advantage_opt[p]))
        for p in range(self._num_players)
    ]

    self._strategy_memories = None
    self._policy_network = MLP(
        self._embedding_size,
        list(policy_network_layers),
        self._num_actions,
        lambda x: x,
        seed,
    )
    self._policy_opt = nn.Optimizer(
        self._policy_network, optax.adam(self._learning_rate), wrt=nn.Param
    )
    self._empty_police_state = nn.state(
        (self._policy_network, self._policy_opt)
    )
    self._policy_graphdef = nn.graphdef(
        (self._policy_network, self._policy_opt)
    )

    # initialise losses
    self._advantage_loss = self._policy_loss = jax.vmap(optax.l2_loss)

    # jit param updates and matched regrets calculations
    self._jittable_matched_regrets = self._get_jittable_matched_regrets()
    self._jittable_adv_update = self._make_train_step(
        self._get_jittable_adv_update()
    )
    self._jittable_policy_update = self._make_train_step(
        self._get_jittable_policy_update()
    )
    self._cached_policy = LRUCache(2**16)

  def _get_jittable_adv_update(self) -> Callable:
    """Get jittable advantage update function."""

    def _loss_adv(
        advantage_model: nn.Module,
        info_states: chex.Array,
        samp_regrets: chex.Array,
        masks: chex.Array,
        iterations: chex.Array,
    ) -> chex.Array:
      """Loss function for the advantage network."""
      preds = forward(advantage_model, info_states)
      preds = jnp.where(masks, preds, jnp.array(0))
      it = jnp.sqrt(iterations)
      loss_values = self._advantage_loss(preds * it, samp_regrets * it)
      return loss_values.mean()

    def update(
        advantage_model: nn.Module,
        optimiser: nn.Optimizer,
        batch: AdvantageMemory,
    ) -> chex.Array:

      main_loss, grads = nn.value_and_grad(_loss_adv)(
          advantage_model,
          batch.info_state,
          batch.advantage,
          batch.legal_mask,
          batch.iteration,
      )
      optimiser.update(advantage_model, grads)
      return main_loss

    return update

  def _get_jittable_policy_update(self) -> Callable:
    """Get jittable policy update function."""

    def _loss_policy(
        policy_model: nn.Module,
        info_states: chex.Array,
        action_probs: chex.Array,
        masks: chex.Array,
        iterations: chex.Array,
    ) -> chex.Array:
      """Loss function for our policy network."""
      preds = forward(policy_model, info_states)
      # masking illegal actions and normalising
      preds = jnp.where(masks, preds, ILLEGAL_ACTION_LOGITS_PENALTY)
      preds = nn.softmax(preds)
      it = jnp.sqrt(iterations)
      loss_values = self._policy_loss(preds * it, action_probs * it)
      return loss_values.mean()

    def update(
        policy_model: nn.Module,
        optimiser: nn.Optimizer,
        batch: StrategyMemory,
    ) -> chex.Array:

      main_loss, grads = nn.value_and_grad(_loss_policy)(
          policy_model,
          batch.info_state,
          batch.strategy_action_probs,
          batch.legal_mask,
          batch.iteration,
      )
      optimiser.update(policy_model, grads)
      return main_loss

    return update

  def _get_jittable_matched_regrets(self) -> Callable:
    """Get jittable regret matching function."""

    @functools.partial(jax.jit, static_argnames=("graphdef",))
    def get_matched_regrets(
        graphdef: nn.GraphDef,
        state: nn.State,
        info_state: chex.Array,
        legal_actions_mask: chex.Array,
    ) -> tuple[chex.Array, chex.Array]:
      advantage_model = nn.merge(graphdef, state)
      advs = advantage_model(info_state)
      advs = jnp.where(legal_actions_mask, advs, ILLEGAL_ACTION_LOGITS_PENALTY)
      advantages = nn.relu(advs)
      summed_regret = jnp.sum(advantages)
      matched_regrets = jnp.where(
          summed_regret > 0,
          advantages / summed_regret,
          jax.nn.one_hot(  # pylint: disable=g-long-lambda
              jnp.argmax(advs), self._num_actions
          ),
      )
      return advantages, matched_regrets

    return get_matched_regrets

  def _next_rng_key(self) -> chex.PRNGKey:
    """Get the next rng subkey from class rngkey."""
    self._rngkey, subkey = jax.random.split(self._rngkey)
    return subkey

  def _reinitialize_policy_network(self) -> None:
    """Reinitalize policy network and optimizer for training."""
    nn.update(
        (self._policy_network, self._policy_opt), self._empty_police_state
    )

  def _reinitialize_advantage_network(self, player: int) -> None:
    """Reinitalize player's advantage network and optimizer for training."""
    nn.update(
        (self._advantage_networks[player], self._advantage_opt[player]),
        self._empty_advantage_states[player],
    )

  @property
  def advantage_buffers(self) -> ReservoirBufferState:
    return self._advantage_memories

  @property
  def strategy_buffer(self) -> ReservoirBufferState:
    return self._strategy_memories

  def _append_to_advantage_buffer(
      self, player: int, data: AdvantageMemory
  ) -> None:
    if self._advantage_memories[player] is None:
      self._advantage_memories[player] = ReservoirBuffer.init(
          self._memory_capacity, data
      )

    self._advantage_memories[player] = ReservoirBuffer.append(
        self._advantage_memories[player], data, self._next_rng_key()
    )

  def _append_to_stategy_buffer(self, data: StrategyMemory) -> None:
    if self._strategy_memories is None:
      self._strategy_memories = ReservoirBuffer.init(
          self._memory_capacity, data
      )

    self._strategy_memories = ReservoirBuffer.append(
        self._strategy_memories, data, self._next_rng_key()
    )

  def solve(self) -> tuple[nn.Module, dict, chex.Numeric]:
    """Solution logic for Deep CFR."""
    advantage_losses = collections.defaultdict(list)
    for _ in range(self._num_iterations):

      if self._print_nash_convs:
        policy_loss = self._learn_strategy_network()
        average_policy = policy.tabular_policy_from_callable(
            self._game, self.action_probabilities
        )
        conv = exploitability.nash_conv(self._game, average_policy)
        print(
            f"NashConv @ {self._iteration} = {conv} | Policy loss ="
            f" {policy_loss}"
        )
        self._reinitialize_policy_network()
        self._cached_policy.clear()

      for p in range(self._num_players):
        for _ in range(self._num_traversals):
          self._traverse_game_tree(self._root_node, p)
        if self._reinitialize_advantage_networks:
          # Re-initialize advantage network for p and train from scratch.
          self._reinitialize_advantage_network(p)
        advantage_losses[p].append(self._learn_advantage_network(p))

      self._iteration += 1

    # Train policy network.
    policy_loss = self._learn_strategy_network()
    return self._policy_network, advantage_losses, policy_loss

  def _traverse_game_tree(self, state, player: int) -> chex.Array:
    """Performs a traversal of the game tree using external sampling.

    Over a traversal the advantage and strategy memories are populated with
    computed advantage values and matched regrets respectively.

    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index for this traversal.

    Returns:
      Recursively returns expected payoffs for each action.
    """
    if state.is_terminal():
      # Terminal state get returns.
      return state.returns()[player]
    elif state.is_chance_node():
      # If this is a chance node, sample an action
      chance_outcome, chance_proba = zip(*state.chance_outcomes())
      action = np.random.choice(
          np.asarray(chance_outcome), p=np.asarray(chance_proba)
      )
      return self._traverse_game_tree(state.child(action), player)
    elif state.current_player() == player:
      # Update the policy over the info set & actions via regret matching.
      _, strategy = self._sample_action_from_advantage(state, player)
      strategy = np.array(strategy)
      exp_payoff = 0 * strategy
      for action in state.legal_actions():
        exp_payoff[action] = self._traverse_game_tree(
            state.child(action), player)
      cfv = np.sum(exp_payoff * strategy)
      samp_regret = (exp_payoff - cfv) * state.legal_actions_mask(player)
      data = AdvantageMemory(
          jnp.asarray(state.information_state_tensor(), dtype=jnp.float32),
          jnp.asarray(self._iteration, dtype=int).reshape(
              1,
          ),
          jnp.asarray(samp_regret, dtype=jnp.float32),
          jnp.asarray(state.legal_actions_mask(player), dtype=jnp.bool),
      )
      self._append_to_advantage_buffer(player, data)
      return cfv
    else:
      other_player = state.current_player()
      _, strategy = self._sample_action_from_advantage(state, other_player)
      # Recompute distribution for numerical errors.
      probs = np.array(strategy)
      probs /= probs.sum()
      sampled_action = np.random.choice(np.arange(self._num_actions), p=probs)

      data = StrategyMemory(
          jnp.asarray(
              state.information_state_tensor(other_player), dtype=jnp.float32
          ),
          jnp.asarray(self._iteration, dtype=int).reshape(
              -1,
          ),
          jnp.asarray(probs, dtype=jnp.float32),
          jnp.asarray(state.legal_actions_mask(other_player), dtype=np.bool),
      )
      self._append_to_stategy_buffer(data)
      return self._traverse_game_tree(state.child(sampled_action), player)

  def _sample_action_from_advantage(
      self, state, player: int
  ) -> tuple[chex.Array, chex.Array]:
    """Returns an info state policy by applying regret-matching.

    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index over which to compute regrets.

    Returns:
      1. (jnp-array) Advantage values for info state actions indexed by action.
      2. (jnp-array) Matched regrets, prob for actions indexed by action.
    """
    self._advantage_networks[player].eval()
    info_state = jnp.asarray(
        state.information_state_tensor(player), dtype=jnp.float32
    )
    legal_actions_mask = jnp.asarray(
        state.legal_actions_mask(player), dtype=jnp.bool
    )

    graphdef, state = nn.split(self._advantage_networks[player])
    advantages, matched_regrets = self._jittable_matched_regrets(
        graphdef, state, info_state, legal_actions_mask
    )

    return advantages, matched_regrets

  def action_probabilities(
      self, state, player_id: int = None
  ) -> dict[chex.Numeric, chex.Array]:
    """Returns action probabilities dict for a single batch."""
    # SLOW METHOD
    del player_id  # unused
    self._policy_network.eval()
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)
    info_state_vector = np.asarray(
        state.information_state_tensor(), dtype=jnp.float32
    )

    def _cached_inference(info_state):
      info_state = jnp.asarray(info_state, dtype=jnp.float32)
      cache_key = info_state.tobytes()
      probs = self._cached_policy.make(
          cache_key, lambda: self._policy_network(info_state)
      )
      return probs

    probs = _cached_inference(info_state_vector)
    legal_actions_mask = jnp.asarray(
        state.legal_actions_mask(cur_player), dtype=jnp.bool
    )
    probs = jnp.where(legal_actions_mask, probs, ILLEGAL_ACTION_LOGITS_PENALTY)
    probs = nn.softmax(probs)

    return {action: probs[action] for action in legal_actions}

  def _make_train_step(
      self,
      update_fn: Callable,
  ) -> Callable:
    """Utility method to avoid boilerplate."""

    @jax.jit
    def _train_step(graphdef, state, batch) -> tuple:

      # merge at the beginning of the function
      model, optimiser = nn.merge(graphdef, state, copy=True)

      loss = update_fn(model, optimiser, batch)

      state = nn.state((model, optimiser))
      return (state, loss)

    return _train_step

  def _learn_advantage_network(self, player: int) -> float:
    """Compute the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Args:
      player: (int) player index.

    Returns:
      The average loss over the advantage network of the last batch.
    """

    self._advantage_networks[player].train()

    # Training loops are inspired by
    # https://flax.readthedocs.io/en/stable/guides/performance.html#functional-training-loop
    state = nn.state(
        (self._advantage_networks[player], self._advantage_opt[player])
    )
    buffer_state = self._advantage_memories[player]
    batch_size = (
        self._batch_size_advantage
        if self._batch_size_advantage
        else len(buffer_state)
    )

    if self._batch_size_advantage is not None:
      if self._batch_size_advantage > len(buffer_state):
        # Not enough samples to train on
        return None

    rng = self._next_rng_key()

    for _ in range(self._advantage_network_train_steps):
      rng, rng_ = jax.random.split(rng)
      batch = ReservoirBuffer.sample(rng_, buffer_state, batch_size)
      state, main_loss = self._jittable_adv_update(
          self._advantage_graphdefs[player], state, batch
      )

    nn.update(
        (self._advantage_networks[player], self._advantage_opt[player]), state
    )

    return main_loss

  def _learn_strategy_network(self) -> float:
    """Compute the loss over the strategy network.

    Returns:
      The average loss obtained on the last training batch of transitions
      or `None`.
    """
    self._policy_network.train()

    if self._strategy_memories is None:
      return None

    if self._batch_size_strategy:
      if self._batch_size_strategy > len(self._strategy_memories):
        ## Skip if there aren't enough samples
        return None

    # Training loops are inspired by
    # https://flax.readthedocs.io/en/stable/guides/performance.html#functional-training-loop
    state = nn.state((self._policy_network, self._policy_opt))
    buffer_state = self._strategy_memories
    batch_size = (
        self._batch_size_strategy
        if self._batch_size_strategy
        else len(buffer_state)
    )

    rng = self._next_rng_key()
    for _ in range(self._advantage_network_train_steps):
      rng, rng_ = jax.random.split(rng)
      batch = ReservoirBuffer.sample(rng_, buffer_state, batch_size)
      state, main_loss = self._jittable_policy_update(
          self._policy_graphdef, state, batch
      )

    nn.update((self._policy_network, self._policy_opt), state)

    return main_loss



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Deep CFR Algorithm.

See https://arxiv.org/abs/1811.00164.

The algorithm defines an `advantage` and `strategy` networks that compute
advantages used to do regret matching across information sets and to approximate
the strategy profiles of the game.  To train these networks a fixed ring buffer
(other data structures may be used) memory is used to accumulate samples to
train the networks.
"""

# pylint: disable=g-explicit-length-test

import collections
from typing import Iterable, NamedTuple

import numpy as np
import torch
from torch import nn
import tree as np_tree

from open_spiel.python import policy
from open_spiel.python.algorithms import exploitability
import pyspiel


def set_seed(seed):
  np.random.seed(seed)
  torch.manual_seed(seed)
  if torch.cuda.is_available():
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


class AdvantageMemory(NamedTuple):
  """Advantage network memory buffer."""

  info_state: np.ndarray
  iteration: np.ndarray
  advantage: np.ndarray


class StrategyMemory(NamedTuple):
  """Stratefy network memory buffer."""

  info_state: np.ndarray
  iteration: np.ndarray
  strategy_action_probs: np.ndarray


class MLP(nn.Module):
  """A simple network built from nn.linear layers."""

  def __init__(
      self,
      input_size: int,
      hidden_sizes: Iterable[int],
      output_size: int,
      final_activation: nn.Module = None,
      seed: int = 42,
  ) -> None:
    """Create the MLP.

    Args:
      input_size: (int) number of inputs
      hidden_sizes: (list) sizes (number of units) of each hidden layer
      output_size: (int) number of outputs
      final_activation: (nn.Module) an activation for the final later, defaults
        to None
      seed: (int) a random seed
    """

    super().__init__()
    set_seed(seed)

    layers_ = []

    def _create_linear_block(in_features, out_features):
      return nn.Sequential(nn.Linear(in_features, out_features), nn.ReLU())

    # Input and Hidden layers
    for size in hidden_sizes:
      layers_.append(_create_linear_block(input_size, size))
      input_size = size
    # Output layer
    layers_.append(nn.LayerNorm(input_size))
    layers_.append(nn.Linear(input_size, output_size))
    if final_activation:
      layers_.append(final_activation)
    self.model = nn.Sequential(*layers_)

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    return self.model(x)

  def reset(self):

    @torch.no_grad()
    def weight_reset(m: nn.Module):
      reset_parameters = getattr(m, "reset_parameters", None)
      if callable(reset_parameters):
        m.reset_parameters()

    self.apply(fn=weight_reset)


class ReservoirBuffer:
  """Allows uniform sampling over a stream of data.

  See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
  """

  def __init__(
      self, capacity: np.ndarray, experience: AdvantageMemory | StrategyMemory
  ) -> None:
    self.capacity = capacity
    self.experience = experience
    self.add_calls = np.array(0)

  def __len__(self) -> int:
    return min(self.add_calls.item(), self.capacity.item())

  def __getitem__(self, idx):
    return np_tree.map_structure(lambda data: data[idx], self.experience)

  @classmethod
  def init(
      cls, capacity: int, experience: AdvantageMemory | StrategyMemory
  ) -> "ReservoirBuffer":
    # Initialize buffer by replicating the structure of the experience
    experience_ = np_tree.map_structure(
        lambda x: np.empty((capacity, *x.shape), dtype=x.dtype), experience
    )
    return cls(np.array(capacity), experience_)

  def append(
      self,
      experience: AdvantageMemory | StrategyMemory,
  ) -> None:
    """Potentially adds `experience` to the reservoir buffer.

    Args:
      experience: data to be added to the reservoir buffer.

    Returns:
      None as the method updated the buffer in-place
    """
    # Determine the insertion index
    # Note: count + 1 because the current item is the (count+1)-th item
    idx = np.random.randint(0, self.add_calls + 1)

    # 2. Logic:
    # If buffer is not full, we always add at 'count'.
    # If buffer is full, we replace at 'idx' ONLY IF idx < capacity.
    is_full = self.add_calls >= self.capacity
    write_idx = np.where(is_full, idx, self.add_calls)
    should_update = write_idx < self.capacity

    def _inplace(arr, idx, val):
      arr[idx] = val

    if should_update:
      np_tree.map_structure(
          lambda buf_leaf, exp_leaf: _inplace(buf_leaf, write_idx, exp_leaf),
          self.experience,
          experience,
      )
    self.add_calls += 1

  def sample(self, num_samples: int) -> AdvantageMemory | StrategyMemory:
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.
    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer
    """
    max_size = len(self)
    if max_size < num_samples:
      raise ValueError(
          f"{num_samples} elements could not be sampled from size {max_size}"
      )

    indices = np.random.choice(max_size, size=(num_samples,), replace=False)

    return np_tree.map_structure(lambda data: data[indices], self.experience)

  def shuffle(self) -> None:
    """Shuffling the reservoir buffer along the batch axis."""
    np_tree.map_structure(
        lambda x: np.random.shuffle(x[: len(self)]), self.experience
    )

  def clear(self) -> None:
    """Clears the reservoir buffer."""
    self.add_calls = np.array(0)


class DeepCFRSolver(policy.Policy):
  """Implements a solver for the Deep CFR Algorithm with PyTorch.

  See https://arxiv.org/abs/1811.00164.

  Define all networks and sampling buffers/memories.  Derive losses & learning
  steps. Initialize the game state and algorithmic variables.

  Note: batch sizes default to `None` implying that training over the full
        dataset in memory is done by default.  To sample from the memories you
        may set these values to something less than the full capacity of the
        memory.
  """

  def __init__(
      self,
      game,
      policy_network_layers=(256, 256),
      advantage_network_layers=(128, 128),
      num_iterations: int = 100,
      num_traversals: int = 20,
      learning_rate: float = 1e-4,
      batch_size_advantage: int = None,
      batch_size_strategy: int = None,
      memory_capacity: int = int(1e6),
      policy_network_train_steps: int = 1,
      advantage_network_train_steps: int = 1,
      reinitialize_advantage_networks: bool = True,
      device: str = "cpu",
      seed: int = 42,
      print_nash_convs: bool = False,
  ) -> None:
    """Initialize the Deep CFR algorithm.

    Args:
      game: OpenSpiel game.
      policy_network_layers: (list[int]) Layer sizes of strategy net MLP.
      advantage_network_layers: (list[int]) Layer sizes of advantage net MLP.
      num_iterations: (int) Number of training iterations.
      num_traversals: (int) Number of traversals per iteration.
      learning_rate: (float) Learning rate.
      batch_size_advantage: (int or None) Batch size to sample from advantage
        memories.
      batch_size_strategy: (int or None) Batch size to sample from strategy
        memories.
      memory_capacity: Number af samples that can be stored in memory.
      policy_network_train_steps: Number of policy network training steps (per
        iteration).
      advantage_network_train_steps: Number of advantage network training steps
        (per iteration).
      reinitialize_advantage_networks: Whether to re-initialize the advantage
        network before training on each iteration.
      device: (str) A pytorch device, defaults to cpu
      seed: (int) A random seed
      print_nash_convs: (bool) print explotability for each iteration, defaults
        to False
    """
    all_players = list(range(game.num_players()))
    super(DeepCFRSolver, self).__init__(game, all_players)

    self._game = game
    if game.get_type().dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS:
      # `_traverse_game_tree` does not take into account this option.
      raise ValueError("Simulatenous games are not supported.")
    self._batch_size_advantage = batch_size_advantage
    self._batch_size_strategy = batch_size_strategy
    self._policy_network_train_steps = policy_network_train_steps
    self._advantage_network_train_steps = advantage_network_train_steps
    self._num_players = game.num_players()
    self._root_node = self._game.new_initial_state()
    self._embedding_size = len(self._root_node.information_state_tensor(0))
    self._num_iterations = num_iterations
    self._num_traversals = num_traversals
    self._memory_capacity = int(memory_capacity)
    self._reinitialize_advantage_networks = reinitialize_advantage_networks
    self._num_actions = game.num_distinct_actions()
    self._iteration = 1
    self._learning_rate = learning_rate
    self._print_nash_convs = print_nash_convs
    self._device = torch.device(device)

    # Define advantage network, loss & memory. (One per player)
    self._advantage_memories = [None] * self._num_players
    self._advantage_networks = [
        MLP(
            self._embedding_size,
            list(advantage_network_layers),
            self._num_actions,
            None,
            seed + p,
        ).to(self._device)
        for p in range(self._num_players)
    ]

    # Define strategy network, loss & memory.
    self._strategy_memories = None
    self._policy_network = MLP(
        self._embedding_size,
        list(policy_network_layers),
        self._num_actions,
        nn.Softmax(-1),
        seed,
    ).to(self._device)

    # Illegal actions are handled in the traversal code where expected payoff
    # and sampled regret is computed from the advantage networks.

    self._loss_policy = nn.MSELoss()
    self._optimizer_policy = None
    self._reinitialize_policy_network()

    self._loss_advantages = nn.MSELoss(reduction="mean")
    self._optimizer_advantages = [None] * self._num_players
    for p in range(self._num_players):
      self._reinitialize_advantage_network(p)

  def _get_buffer_init(
      self, capacity: int, data: AdvantageMemory | StrategyMemory
  ) -> ReservoirBuffer:
    return ReservoirBuffer.init(capacity, data)

  @property
  def advantage_buffers(self):
    return self._advantage_memories

  @property
  def strategy_buffer(self):
    return self._strategy_memories

  def _reinitialize_advantage_network(self, player):
    self._advantage_networks[player].reset()
    self._optimizer_advantages[player] = torch.optim.Adam(
        self._advantage_networks[player].parameters(), lr=self._learning_rate)

  def _reinitialize_policy_network(self):
    self._policy_network.reset()
    self._optimizer_policy = torch.optim.Adam(
        self._policy_network.parameters(), lr=self._learning_rate
    )

  def solve(self) -> tuple[nn.Module, list[list[float]], float]:
    """Solution logic for Deep CFR.

    Traverses the game tree, while storing the transitions for training
    advantage and policy networks.

    Returns:
      1. (nn.Module) Instance of the trained policy network for inference.
      2. (list of floats) Advantage network losses for
         each player during each iteration.
      3. (float) Policy loss.
    """
    advantage_losses = collections.defaultdict(list)
    for _ in range(self._num_iterations):
      if self._print_nash_convs:
        policy_loss = self._learn_strategy_network()
        average_policy = policy.tabular_policy_from_callable(
            self._game, self.action_probabilities
        )
        conv = exploitability.nash_conv(self._game, average_policy)
        print(
            f"NashConv @ {self._iteration} = {conv} | Policy loss ="
            f" {policy_loss}"
        )
        self._reinitialize_policy_network()

      for p in range(self._num_players):
        for _ in range(self._num_traversals):
          self._traverse_game_tree(self._root_node, p)
        if self._reinitialize_advantage_networks:
          # Re-initialize advantage network for player and train from scratch.
          self._reinitialize_advantage_network(p)
        # Re-initialize advantage networks and train from scratch.
        advantage_losses[p].append(self._learn_advantage_network(p))
      self._iteration += 1
    # Train policy network.
    policy_loss = self._learn_strategy_network()
    return self._policy_network, advantage_losses, policy_loss

  def _append_to_stategy_buffer(self, data: StrategyMemory) -> None:
    if self._strategy_memories is None:
      self._strategy_memories = self._get_buffer_init(
          self._memory_capacity, data
      )

    self._strategy_memories.append(data)

  def _append_to_advantage_buffer(
      self, player: int, data: AdvantageMemory
  ) -> None:
    if self._advantage_memories[player] is None:
      self._advantage_memories[player] = self._get_buffer_init(
          self._memory_capacity, data
      )

    self._advantage_memories[player].append(data)

  def _traverse_game_tree(self, state, player):
    """Performs a traversal of the game tree.

    Over a traversal the advantage and strategy memories are populated with
    computed advantage values and matched regrets respectively.

    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index for this traversal.

    Returns:
      (float) Recursively returns expected payoffs for each action.
    """
    expected_payoff = collections.defaultdict(float)
    if state.is_terminal():
      # Terminal state get returns.
      return state.returns()[player]
    elif state.is_chance_node():
      # If this is a chance node, sample an action
      chance_outcome, chance_proba = zip(*state.chance_outcomes())
      action = np.random.choice(chance_outcome, p=chance_proba)
      return self._traverse_game_tree(state.child(action), player)
    elif state.current_player() == player:
      sampled_regret = collections.defaultdict(float)
      # Update the policy over the info set & actions via regret matching.
      _, strategy = self._sample_action_from_advantage(state, player)
      for action in state.legal_actions():
        expected_payoff[action] = self._traverse_game_tree(
            state.child(action), player)
      cfv = 0
      for a_ in state.legal_actions():
        cfv += strategy[a_] * expected_payoff[a_]
      for action in state.legal_actions():
        sampled_regret[action] = expected_payoff[action]
        sampled_regret[action] -= cfv
      sampled_regret_arr = [0] * self._num_actions
      for action in sampled_regret:
        sampled_regret_arr[action] = sampled_regret[action]

      data = AdvantageMemory(
          np.array(state.information_state_tensor(), dtype=np.float32),
          np.array(self._iteration, dtype=int).reshape(
              1,
          ),
          np.array(sampled_regret_arr, dtype=np.float32),
      )

      self._append_to_advantage_buffer(player, data)
      return cfv
    else:
      other_player = state.current_player()
      _, strategy = self._sample_action_from_advantage(state, other_player)
      # Recompute distribution for numerical errors.
      probs = np.array(strategy)
      probs /= probs.sum()
      sampled_action = np.random.choice(range(self._num_actions), p=probs)

      data = StrategyMemory(
          np.array(
              state.information_state_tensor(other_player), dtype=np.float32
          ),
          np.array(self._iteration, dtype=int).reshape(
              1,
          ),
          np.array(strategy, dtype=np.float32),
      )
      self._append_to_stategy_buffer(data)

      return self._traverse_game_tree(state.child(sampled_action), player)

  @torch.inference_mode()
  def _sample_action_from_advantage(self, state, player):
    """Returns an info state policy by applying regret-matching.

    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index over which to compute regrets.

    Returns:
      1. (list) Advantage values for info state actions indexed by action.
      2. (list) Matched regrets, prob for actions indexed by action.
    """
    info_state = state.information_state_tensor(player)
    legal_actions = state.legal_actions(player)
    with torch.no_grad():
      state_tensor = torch.FloatTensor(
          np.expand_dims(info_state, axis=0), device=self._device
      )
      raw_advantages = (
          self._advantage_networks[player](state_tensor)[0].cpu().numpy()
      )
    advantages = [max(0., advantage) for advantage in raw_advantages]
    cumulative_regret = np.sum([advantages[action] for action in legal_actions])
    matched_regrets = np.array([0.] * self._num_actions)
    if cumulative_regret > 0.:
      for action in legal_actions:
        matched_regrets[action] = advantages[action] / cumulative_regret
    else:
      matched_regrets[max(legal_actions, key=lambda a: raw_advantages[a])] = 1
    return advantages, matched_regrets

  @torch.inference_mode()
  def action_probabilities(self, state, player_id=None):
    """Computes action probabilities for the current player in state.

    Args:
      state: (pyspiel.State) The state to compute probabilities for.
      player_id: unused, but needed to implement the Policy API.

    Returns:
      (dict) action probabilities for a single batch.
    """
    del player_id
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)
    info_state_vector = np.array(state.information_state_tensor())
    if len(info_state_vector.shape) == 1:
      info_state_vector = np.expand_dims(info_state_vector, axis=0)
    probs = (
        self._policy_network(
            torch.FloatTensor(info_state_vector, device=self._device)
        )
        .cpu()
        .numpy()
    )
    return {action: probs[0][action] for action in legal_actions}

  def _learn_advantage_network(self, player):
    """Compute the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Args:
      player: (int) player index.

    Returns:
      (float) The average loss over the advantage network.
    """
    for _ in range(self._advantage_network_train_steps):

      if self._batch_size_advantage:
        if self._batch_size_advantage > len(self._advantage_memories[player]):
          ## Skip if there aren't enough samples
          return None
        samples = self._advantage_memories[player].sample(
            self._batch_size_advantage)
      else:
        self._advantage_memories[player].shuffle()
        samples = self._advantage_memories[player].experience

      # Ensure some samples have been gathered.
      if len(samples.info_state == 0):
        return None

      self._optimizer_advantages[player].zero_grad()
      iters = torch.FloatTensor(samples.iteration, device=self._device).sqrt()
      outputs = self._advantage_networks[player](
          torch.FloatTensor(samples.info_state, device=self._device)
      )
      advantages = torch.FloatTensor(samples.advantage, device=self._device)
      loss_advantages = self._loss_advantages(
          iters * outputs, iters * advantages
      )
      loss_advantages.backward()
      self._optimizer_advantages[player].step()

    return loss_advantages.detach().cpu().item()

  def _learn_strategy_network(self):
    """Compute the loss over the strategy network.

    Returns:
      (float) The average loss obtained on this batch of transitions or `None`.
    """
    if self._strategy_memories is None:
      return None

    for _ in range(self._policy_network_train_steps):
      if self._batch_size_strategy:
        if self._batch_size_strategy > len(self._strategy_memories):
          ## Skip if there aren't enough samples
          return None
        samples = self._strategy_memories.sample(self._batch_size_strategy)
      else:
        self._strategy_memories.shuffle()
        samples = self._strategy_memories.experience

      # Ensure some samples have been gathered.

      if len(samples.info_state) == 0:
        return None

      self._optimizer_policy.zero_grad()
      iters = torch.FloatTensor(samples.iteration, device=self._device).sqrt()
      outputs = self._policy_network(
          torch.FloatTensor(samples.info_state, device=self._device)
      )
      ac_probs = torch.FloatTensor(
          samples.strategy_action_probs, device=self._device
      ).squeeze()
      loss_strategy = self._loss_policy(iters * outputs, iters * ac_probs)
      loss_strategy.backward()
      self._optimizer_policy.step()

    return loss_strategy.detach().cpu().item()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the game-specific functions for go."""

import json

from absl.testing import absltest

import pyspiel

go = pyspiel.go

# First game of Kisei collection, taken from here:
# https://webdocs.cs.ualberta.ca/~mmueller/go/games.html
#
# Nicely formatted game with root node and a single variation and one node per
# mode.
EXAMPLE_SGF_STRING = """
(;GM[1]RE[B+R]PW[Hashimoto Utaro]PB[Fujisawa Shuko]DT[1976-12-02,03]SZ[19]KM[5.5]ID[ 1/1]FF[3];B[pd];W[cq];B[pq];W[po];B[dd];W[oq];B[or];W[op];B[nq]
;W[nr];B[mr];W[pr];B[ns];W[qq];B[mo];W[jc];B[qm];W[mn];B[nn]
;W[pm];B[pl];W[om];B[nm];W[ol];B[pk];W[qi];B[ok];W[nl];B[mm]
;W[qf];B[qe];W[pf];B[nd];W[gc];B[eq];W[do];B[gp];W[cf];B[cm]
;W[cn];B[dm];W[fo];B[ci];W[go];B[hp];W[ee];B[ed];W[bd];B[gd]
;W[fd];B[fe];W[fc];B[de];W[ch];B[ef];W[di];B[cj];W[dh];B[dj]
;W[rk];B[rl];W[qk];B[ql];W[ge];B[gf];W[hd];B[pi];W[fg];B[cc]
;W[qh];B[re];W[ml];B[ph];W[rf];B[lj];W[nf];B[ro];W[rp];B[qo]
;W[kl];B[lm];W[ll];B[bc];W[oc];B[pc];W[nh];B[qp];W[pp];B[rq]
;W[rr];B[sp];W[od];B[ob];W[nb];B[oe];W[nc];B[pe];W[ne];B[of]
;W[og];B[pg];W[qg];B[pb];W[se];B[md];W[lb];B[sd];W[sf];B[rc]
;W[lg];B[kg];W[kh];B[jh];W[kf];B[jg];W[lh];B[fi];W[fh];B[gi]
;W[hf];B[cr];W[dq];B[dr];W[ep];B[fq];W[gm];B[hh];W[bg];B[ff]
;W[hg];B[ei];W[jo];B[jn];W[io];B[jm];W[jl];B[il];W[in];B[ik]
;W[ko];B[dg];W[bi];B[be];W[eh];B[gg];W[bj];B[bk];W[gh];B[aj]
;W[hi];B[ih];W[ej];B[hj])
"""

# Example with two variations within a entry (one root node and two
# variations), where each variation starts with '(' and end with ')'.
EXAMPLE_SGF_STRING_2 = """
(;FF[4]GM[1]CA[UTF-8]AP[besogo:0.0.2-alpha]SZ[9]ST[0]

(;B[de]
;W[ee])
(;B[cd]
;AB[ce][cf][dd][df][ed][ef][fd][ff][gd][ge][gf]AW[bd][be][bf][cc][cg][dc][dg][ec][eg][fc][fg][gc][gg][hd][he][hf]))
"""

# Example with AB and AW properties directly in the root node and without a
# prepended ';'.
EXAMPLE_SGF_STRING_3 = """
(;FF[4]GM[1]CA[UTF-8]AP[besogo:0.0.2-alpha]SZ[13]ST[0]
AB[ik][il][jd][jk][kj][kl][lj]AW[dd][dj][jj])
"""


class GamesGoTest(absltest.TestCase):

  def test_json(self):
    game = pyspiel.load_game("go")
    state = game.new_initial_state()
    state_struct = state.to_struct()
    self.assertEqual(state_struct.current_player, "B")
    self.assertEqual(state_struct.board_grid[0][0]["a1"], "EMPTY")
    state.apply_action(0)
    state_json = state.to_json()
    state_dict = json.loads(state_json)
    state_struct = go.GoStateStruct(state_json)
    self.assertEqual(state_struct.current_player, "W")
    self.assertEqual(state_dict["current_player"], "W")
    self.assertEqual(state_struct.board_grid[0][0]["a1"], "B")
    self.assertEqual(state_struct.to_json(),
                     json.dumps(state_dict, separators=(",", ":")))

  def test_games_load_from_sgf_string(self):
    games_and_states = go.load_games_from_sgf_string(EXAMPLE_SGF_STRING)
    self.assertLen(games_and_states, 1)
    pyspiel.random_sim_test_with_specific_initial_state(
        games_and_states[0][0], 1, games_and_states[0][1], serialize=True
    )

    games_and_states = go.load_games_from_sgf_string(EXAMPLE_SGF_STRING_2)
    self.assertLen(games_and_states, 2)
    pyspiel.random_sim_test_with_specific_initial_state(
        games_and_states[0][0], 1, games_and_states[0][1], serialize=True
    )
    pyspiel.random_sim_test_with_specific_initial_state(
        games_and_states[1][0], 1, games_and_states[1][1], serialize=False
    )

    games_and_states = go.load_games_from_sgf_string(EXAMPLE_SGF_STRING_3)
    self.assertLen(games_and_states, 1)
    pyspiel.random_sim_test_with_specific_initial_state(
        games_and_states[0][0], 1, games_and_states[0][1], serialize=False
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the game-specific functions for connect_four."""

from typing import Any

from absl.testing import absltest
from absl.testing import parameterized
import pyspiel

connect_four = pyspiel.connect_four


def empty_board():
  """Returns an empty 6x7 board."""
  return [["." for _ in range(7)] for _ in range(6)]


class GamesConnectFourTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.game = pyspiel.load_game("connect_four")

  def test_json(self):
    state = self.game.new_initial_state()
    state.apply_action(4)
    state_struct = state.to_struct()
    self.assertEqual(
        state_struct.board[0],
        [".", ".", ".", ".", "x", ".", "."],
    )
    self.assertEqual(state_struct.current_player, "o")
    json_from_struct = state_struct.to_json()
    state_json = state.to_json()
    self.assertEqual(json_from_struct, state_json)
    state_struct = connect_four.ConnectFourStateStruct(state_json)
    self.assertEqual(state_struct.to_json(), state_json)

  def test_action_struct(self):
    """Test ActionStruct creation and round-trip."""
    state = self.game.new_initial_state()
    for col in range(7):
      action_struct = state.action_to_struct(0, col)
      self.assertIsInstance(action_struct, connect_four.ConnectFourActionStruct)
      self.assertEqual(action_struct.column, col)

      # Test round-trip
      actions = state.struct_to_actions(action_struct)
      self.assertEqual(actions, [col])

  def test_action_struct_json(self):
    """Test ActionStruct JSON serialization."""
    action_struct = connect_four.ConnectFourActionStruct()
    action_struct.column = 3
    json_str = action_struct.to_json()
    self.assertEqual(json_str, '{"column":3}')

    # Parse back
    parsed = connect_four.ConnectFourActionStruct(json_str)
    self.assertEqual(parsed.column, 3)

  def test_state_from_dict_basic(self):
    """Test creating state from dictionary."""
    state = self.game.new_initial_state()
    state.apply_action(3)
    state.apply_action(4)

    # Round-trip through dict
    new_state = self.game.new_initial_state(state.to_dict())
    self.assertEqual(new_state.to_dict(), state.to_dict())
    self.assertEqual(str(new_state), str(state))

  def test_state_from_dict_terminal_x_wins(self):
    """Test creating terminal state where X wins."""
    board = empty_board()
    board[0][3] = "x"
    board[1][3] = "x"
    board[2][3] = "x"
    board[3][3] = "x"
    board[0][4] = "o"
    board[1][4] = "o"
    board[2][4] = "o"

    state_dict = {
        "board": board,
        "current_player": "Terminal",
        "is_terminal": True,
        "winner": "x",
    }

    state = self.game.new_initial_state(state_dict)
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns(), [1.0, -1.0])

  def test_state_from_dict_terminal_draw(self):
    """Test creating terminal draw state."""
    # A full board with no winner
    board = [
        ["o", "o", "x", "x", "x", "o", "o"],
        ["x", "x", "o", "o", "o", "x", "x"],
        ["o", "o", "x", "x", "x", "o", "o"],
        ["x", "x", "o", "o", "o", "x", "x"],
        ["o", "o", "x", "x", "x", "o", "o"],
        ["x", "x", "o", "o", "o", "x", "x"],
    ]

    state_dict = {
        "board": board,
        "current_player": "Terminal",
        "is_terminal": True,
        "winner": "draw",
    }

    state = self.game.new_initial_state(state_dict)
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns(), [0.0, 0.0])

  def test_state_from_json(self):
    """Test creating state from JSON string."""
    state = self.game.new_initial_state()
    state.apply_action(3)
    state.apply_action(4)

    json_str = state.to_json()
    new_state = self.game.new_initial_state(json_str)
    self.assertEqual(new_state.to_json(), json_str)

  @parameterized.parameters(
      # Gap in column
      dict(
          invalid_dict={
              "board": [
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", "x", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "o",
              "is_terminal": False,
              "winner": "",
          },
          error_msg="gap",
      ),
      # Invalid cell value
      dict(
          invalid_dict={
              "board": [
                  [".", ".", ".", "X", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "o",
              "is_terminal": False,
              "winner": "",
          },
          error_msg="Invalid cell",
      ),
      # is_terminal mismatch
      dict(
          invalid_dict={
              "board": [
                  [".", ".", ".", "x", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "Terminal",
              "is_terminal": True,
              "winner": "x",
          },
          error_msg="is_terminal",
      ),
      # winner mismatch
      dict(
          invalid_dict={
              "board": [
                  [".", ".", ".", "x", "o", ".", "."],
                  [".", ".", ".", "x", "o", ".", "."],
                  [".", ".", ".", "x", "o", ".", "."],
                  [".", ".", ".", "x", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "Terminal",
              "is_terminal": True,
              "winner": "o",
          },
          error_msg="winner",
      ),
      # Piece count imbalance (strict mode)
      dict(
          invalid_dict={
              "board": [
                  ["x", "x", "x", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "o",
              "is_terminal": False,
              "winner": "",
          },
          error_msg="imbalance",
      ),
      # current_player doesn't match piece counts (strict mode)
      dict(
          invalid_dict={
              "board": [
                  ["x", "o", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "o",
              "is_terminal": False,
              "winner": "",
          },
          error_msg="current_player",
      ),
      # Wrong board dimensions (5 columns instead of 7)
      dict(
          invalid_dict={
              "board": [
                  [".", ".", ".", ".", "."],
                  [".", ".", ".", ".", "."],
                  [".", ".", ".", ".", "."],
                  [".", ".", ".", ".", "."],
                  [".", ".", ".", ".", "."],
                  [".", ".", ".", ".", "."],
              ],
              "current_player": "x",
              "is_terminal": False,
              "winner": "",
          },
          error_msg="column count",
      ),
      # Both players have winning lines
      dict(
          invalid_dict={
              "board": [
                  ["x", "o", ".", ".", ".", ".", "."],
                  ["x", "o", ".", ".", ".", ".", "."],
                  ["x", "o", ".", ".", ".", ".", "."],
                  ["x", "o", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
                  [".", ".", ".", ".", ".", ".", "."],
              ],
              "current_player": "Terminal",
              "is_terminal": True,
              "winner": "x",
          },
          error_msg="both",
      ),
  )
  def test_state_from_invalid_dict(self, invalid_dict: dict[str, Any],
                                   error_msg: str):
    with self.assertRaises(pyspiel.SpielError) as context:
      self.game.new_initial_state(invalid_dict)
    self.assertIn(error_msg.lower(), str(context.exception).lower())

  def test_starting_state(self):
    """Test that starting_state is preserved."""
    state = self.game.new_initial_state()
    state.apply_action(3)
    state.apply_action(4)

    # Create state from dict
    new_state = self.game.new_initial_state(state.to_dict())
    self.assertIsNotNone(new_state.starting_state())
    self.assertEqual(str(new_state.starting_state()), str(state))

    # Apply more actions and verify starting_state unchanged
    new_state.apply_action(5)
    self.assertEqual(str(new_state.starting_state()), str(state))

  def test_starting_state_after_clone(self):
    """Test that starting_state is preserved through clone."""
    state = self.game.new_initial_state()
    state.apply_action(3)

    new_state = self.game.new_initial_state(state.to_dict())
    new_state.apply_action(4)
    clone = new_state.clone()

    self.assertEqual(str(clone.starting_state()), str(state))

  def test_game_params_default(self):
    """Test GameParams with default values."""
    params = connect_four.ConnectFourGameParams()
    self.assertEqual(params.game_name, "connect_four")
    self.assertEqual(params.rows, 6)
    self.assertEqual(params.columns, 7)
    self.assertEqual(params.x_in_row, 4)
    self.assertEqual(params.egocentric_obs_tensor, False)

    game = pyspiel.load_game(params)
    self.assertEqual(game.get_type().short_name, "connect_four")

  def test_game_params_custom(self):
    """Test GameParams with custom values."""
    params = connect_four.ConnectFourGameParams()
    params.rows = 8
    params.columns = 9
    params.x_in_row = 5

    game = pyspiel.load_game(params)
    state = game.new_initial_state()

    # Verify we can play on the larger board
    self.assertLen(state.legal_actions(), 9)

    # Fill a column to verify row count
    for _ in range(8):
      if state.is_terminal():
        break
      state.apply_action(0)
      if not state.is_terminal():
        state.apply_action(1)
    # Either column 0 is full or the game ended

  def test_game_params_json(self):
    """Test GameParams JSON serialization."""
    params = connect_four.ConnectFourGameParams()
    params.rows = 5
    params.columns = 6
    json_str = params.to_json()

    self.assertIn('"rows":5', json_str)
    self.assertIn('"columns":6', json_str)
    self.assertIn('"game_name":"connect_four"', json_str)

  def test_load_game_from_json(self):
    """Test loading game from JSON string."""
    game = pyspiel.load_game_from_json(
        '{"game_name":"connect_four","rows":4,"columns":5,"x_in_row":3}')
    self.assertEqual(game.get_type().short_name, "connect_four")

    # Verify custom connect-3 win condition on smaller board
    state = game.new_initial_state()
    self.assertLen(state.legal_actions(), 5)

    # Win with 3 in a row
    state.apply_action(0)  # x
    state.apply_action(1)  # o
    state.apply_action(0)  # x
    state.apply_action(1)  # o
    state.apply_action(0)  # x - wins with connect-3
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.returns(), [1.0, -1.0])


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A helpful wrapper for the Elo rating system."""

import numpy as np
from open_spiel.python.algorithms import elo
import pyspiel
from open_spiel.python.voting import base


def compute_ratings_from_preference_profile(
    profile: base.PreferenceProfile,
    smoothing_factor: float = pyspiel.elo.DEFAULT_SMOOTHING_FACTOR,
    max_iterations: int = pyspiel.elo.DEFAULT_MAX_ITERATIONS,
    convergence_delta: float = pyspiel.elo.DEFAULT_CONVERGENCE_DELTA,
) -> dict[base.AlternativeId, float]:
  """Compute Elo ratings from a win matrix and a draw matrix."""
  options = pyspiel.elo.default_elo_options()
  options.smoothing_factor = smoothing_factor
  options.max_iterations = max_iterations
  options.convergence_delta = convergence_delta
  num_agents = profile.num_alternatives()
  alt_idx = profile.alternatives_dict
  win_matrix = np.zeros((num_agents, num_agents), dtype=int)
  for vote in profile.votes:
    for i in range(len(vote.vote)):
      for j in range(i + 1, len(vote.vote)):
        for _ in range(vote.weight):
          agent_i_idx = alt_idx[vote.vote[i]]
          agent_j_idx = alt_idx[vote.vote[j]]
          win_matrix[agent_i_idx, agent_j_idx] += 1
  ratings_array = elo.compute_ratings_from_matrices(
      win_matrix, smoothing_factor=smoothing_factor,
      max_iterations=max_iterations,
      convergence_delta=convergence_delta,
  )
  alternatives = profile.alternatives
  return {alternatives[i]: ratings_array[i] for i in range(num_agents)}


class EloVoting(base.AbstractVotingMethod):
  """Implements Elo as a voting method.

  Note: if there is no data on an alternatives, it is assigned a rating of 0.
  """

  def __init__(
      self,
      smoothing_factor: float = pyspiel.elo.DEFAULT_SMOOTHING_FACTOR,
      max_iterations: int = pyspiel.elo.DEFAULT_MAX_ITERATIONS,
      convergence_delta: float = pyspiel.elo.DEFAULT_CONVERGENCE_DELTA):
    self._smoothing_factor = smoothing_factor
    self._max_iterations = max_iterations
    self._convergence_delta = convergence_delta

  def name(self) -> str:
    return "elo"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    ratings_dict = compute_ratings_from_preference_profile(
        profile, self._smoothing_factor, self._max_iterations,
        self._convergence_delta,
    )
    sorted_ratings = sorted(ratings_dict.items(), key=lambda item: item[1],
                            reverse=True)
    if len(sorted_ratings) < profile.num_alternatives():
      # If missing alternatives, fill them in with zero scores.
      for alt in profile.alternatives:
        if alt not in ratings_dict:
          sorted_ratings.append((alt, 0.0))
    assert len(sorted_ratings) == profile.num_alternatives()
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_ratings)
    return outcome



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Any

from absl.testing import absltest
from absl.testing import parameterized
from flax import linen
from flax import nnx
import jax.numpy as jnp
import numpy as np

from open_spiel.python.algorithms.alpha_zero import utils
import pyspiel

solved = {}


def solve_game(state):
  state_str = str(state)
  if state_str in solved:
    return solved[state_str].value
  if state.is_terminal():
    return state.returns()[0]

  max_player = state.current_player() == 0
  obs = state.observation_tensor()
  act_mask = np.array(state.legal_actions_mask())
  values = np.full(act_mask.shape, -2 if max_player else 2)
  for action in state.legal_actions():
    values[action] = solve_game(state.child(action))
  value = values.max() if max_player else values.min()
  best_actions = np.where((values == value) & act_mask)
  policy = np.zeros_like(act_mask)
  policy[best_actions[0][0]] = 1  # Choose the first for a deterministic policy.
  solved[state_str] = utils.TrainInput(
      observation=jnp.asarray(obs, dtype=jnp.float32),
      legals_mask=jnp.asarray(act_mask, dtype=jnp.bool),
      policy=jnp.asarray(policy, dtype=jnp.float32),
      value=jnp.asarray(value, dtype=jnp.float32),
  )
  return value


def build_model(game: Any, api_version: str, model_type: str):
  return utils.api_selector(api_version).Model.build_model(
      model_type,
      game.observation_tensor_shape(),
      game.num_distinct_actions(),
      nn_width=32,
      nn_depth=2,
      weight_decay=1e-4,
      learning_rate=0.01,
      path=None,
  )


class ModelTest(parameterized.TestCase):

  @parameterized.parameters(
      itertools.product(
          utils.AVIALABLE_APIS,
          utils.api_selector(utils.AVIALABLE_APIS[0]).Model.valid_model_types,
      )
  )
  def test_model_learns_simple(self, api_version: str, model_type: str):
    game = pyspiel.load_game("tic_tac_toe")
    model = build_model(game, api_version, model_type)
    print("Num variables:", model.num_trainable_variables)
    model.print_trainable_variables()

    train_inputs = []
    state = game.new_initial_state()
    while not state.is_terminal():
      obs = state.observation_tensor()
      act_mask = state.legal_actions_mask()
      action = state.legal_actions()[0]
      policy = np.zeros(len(act_mask), dtype=float)
      policy[action] = 1
      train_inputs.append(
          utils.TrainInput(
              observation=jnp.asarray(obs),
              legals_mask=jnp.asarray(act_mask),
              policy=jnp.asarray(policy),
              value=jnp.asarray(1),
          )
      )
      state.apply_action(action)
      value, policy = model.inference(obs, act_mask)
      self.assertLen(policy, game.num_distinct_actions())
      self.assertEqual(value.ndim, 0)  # value is a scalar

    losses = []
    policy_loss_goal = 0.05
    value_loss_goal = 0.05

    for i in range(200):
      loss = model.update(utils.TrainInput.stack(train_inputs))
      print(i, loss)
      losses.append(loss)
      if loss.policy < policy_loss_goal and loss.value < value_loss_goal:
        break

    self.assertGreater(losses[0].total, losses[-1].total)
    self.assertGreater(losses[0].policy, losses[-1].policy)
    self.assertGreater(losses[0].value, losses[-1].value)
    self.assertLess(losses[-1].value, value_loss_goal)
    self.assertLess(losses[-1].policy, policy_loss_goal)

  @absltest.skip("Too slow for the CI, because it takes much time.")
  @parameterized.parameters(
      itertools.product(
          utils.AVIALABLE_APIS,
          utils.api_selector(utils.AVIALABLE_APIS[0]).Model.valid_model_types,
      )
  )
  def test_model_learns_optimal(self, api_version: str, model_type: str):
    game = pyspiel.load_game("tic_tac_toe")
    solve_game(game.new_initial_state())

    model = build_model(game, api_version, model_type)
    print("Num variables:", model.num_trainable_variables)
    model.print_trainable_variables()

    train_inputs = list(solved.values())
    print("states:", len(train_inputs))
    losses = []
    policy_loss_goal = 0.12
    value_loss_goal = 0.12
    for i in range(500):
      loss = model.update(utils.TrainInput.stack(train_inputs))
      print(i, loss)
      losses.append(loss)
      if loss.policy < policy_loss_goal and loss.value < value_loss_goal:
        break

    self.assertGreater(losses[0].policy, losses[-1].policy)
    self.assertGreater(losses[0].value, losses[-1].value)
    self.assertGreater(losses[0].total, losses[-1].total)
    self.assertLess(losses[-1].value, value_loss_goal)
    self.assertLess(losses[-1].policy, policy_loss_goal)

  @absltest.skip("May save to the disk")
  @parameterized.parameters(
      itertools.product(
          utils.AVIALABLE_APIS,
          utils.api_selector(utils.AVIALABLE_APIS[0]).Model.valid_model_types,
      )
  )
  def test_conversions(self, api_version: str, model_type: str):
    game = pyspiel.load_game("tic_tac_toe")
    model = build_model(game, api_version, model_type)

    if isinstance(model, linen.Module):
      nnx_model = build_model(game, "nnx", model_type)
      nnx_model._state = utils.api_selector("linen").Model._create_train_state(
          utils.linen_to_nnx(model._model), model._state.tx
      )
      model = nnx_model
    elif isinstance(model, nnx.Module):
      config = {
          "model_type": model_type,
          "input_shape": game.observation_tensor_shape(),
          "output_size": game.num_distinct_actions(),
          "nn_width": 32,
          "nn_depth": 2,
      }
      apply_fn, variables = utils.nnx_to_linen(
          model._model, game.observation_tensor_shape(), **config
      )
      linen_model = build_model(game, "linen", model_type)
      linen_model._state = utils.api_selector("nnx").Model._create_train_state(
          apply_fn, variables, model._state.tx
      )
      model = linen_model

    state = game.new_initial_state()
    while not state.is_terminal():
      obs = state.observation_tensor()
      act_mask = state.legal_actions_mask()
      action = state.legal_actions()[0]
      policy = np.zeros(len(act_mask), dtype=float)
      policy[action] = 1
      state.apply_action(action)
      value, policy = model.inference(obs, act_mask)
      self.assertLen(policy, game.num_distinct_actions())
      self.assertEqual(value.ndim, 0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2026 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python ESCHER example."""

# pylint: disable=logging-fstring-interpolation

from absl import app
from absl import flags
from absl import logging

from open_spiel.python import policy
from open_spiel.python.algorithms import expected_game_score
import pyspiel
from open_spiel.python.pytorch import escher


FLAGS = flags.FLAGS
flags.DEFINE_string("game_name", "kuhn_poker", "Name of the game")


def main(unused_argv):
  logging.info(f"Loading {FLAGS.game_name}")
  game = pyspiel.load_game(FLAGS.game_name)

  agent = escher.Agent(game, escher.Config())
  train_cfg = escher.TrainConfig(game)
  train_cfg.iterations = 100
  train_cfg.evaluation_interval = 20
  train_cfg.nashconv = True

  escher.train(train_cfg, agent)

  average_policy = policy.tabular_policy_from_callable(
      game, lambda s: _action_probabilities(agent, s)
  )
  pyspiel_policy = policy.python_policy_to_pyspiel_policy(average_policy)
  conv = pyspiel.nash_conv(game, pyspiel_policy)
  logging.info(f"ESCHER in {FLAGS.game_name} - NashConv: {conv}")

  avg_policy_vals = expected_game_score.policy_value(
      game.new_initial_state(), [average_policy] * 2
  )
  if FLAGS.game_name == "kuhn_poker":
    # We know EVs
    logging.info(
        f"Computed player 0 value: {avg_policy_vals[0]:.2f} (expected:"
        f" {-1/18:.2f})."
    )
    logging.info(
        f"Computed player 1 value: {avg_policy_vals[1]:.2f} (expected:"
        f" {1/18:.2f})."
    )
  else:
    logging.info(f"Computed player 0 value: {avg_policy_vals[0]:.2f}")
    logging.info(f"Computed player 1 value: {avg_policy_vals[1]:.2f}")


def _action_probabilities(agent, state):
  probs = agent.action_probabilities(state)

  prob_dict = {}
  for a, m in enumerate(state.legal_actions_mask()):
    if m == 1:
      prob_dict[a] = probs[a]
  return prob_dict


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Games implemented in Python.

These games are registered as they are imported. It's perfectly possible to
import just a single game if you prefer. There is no need to add new games here,
so long as they register themselves and you import them when wanting to use
them. However, adding them here will make them available for playthroughs and
for automated API testing.

Registration looks like this:
```
pyspiel.register_game(_GAME_TYPE, KuhnPokerGame)
```
"""

from open_spiel.python.games import ant_foraging
from open_spiel.python.games import block_dominoes
from open_spiel.python.games import chat_game
from open_spiel.python.games import dynamic_routing
from open_spiel.python.games import hangman
from open_spiel.python.games import iterated_prisoners_dilemma
from open_spiel.python.games import kuhn_poker
from open_spiel.python.games import liars_poker
from open_spiel.python.games import team_dominoes
from open_spiel.python.games import tic_tac_toe

# Optional game modules with additional dependencies
try:
  # pylint: disable=g-import-not-at-top
  from open_spiel.python.games import pokerkit_wrapper
  # repeated_pokerkit depends on pokerkit_wrapper; if the former cannot be
  # imported there is no point importing the latter.
  from open_spiel.python.games import repeated_pokerkit
  # pylint: enable=g-import-not-at-top
except ImportError as e:
  # Initialize to None on failure to ensure that this won't trigger NameError
  # later if someone tries to check for the module's presence.
  print(f"Optional module pokerkit_wrapper was not importable: {e}")
  pokerkit_wrapper = None
  repeated_pokerkit = None
  pass
